# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

from hashlib import sha256

import pytest

from trezorlib import cosi
from trezorlib.tools import parse_path

from .common import TrezorTest


@pytest.mark.skip_t2
class TestCosi(TrezorTest):
    def test_cosi_commit(self):
        self.setup_mnemonic_pin_passphrase()

        digest = sha256(b"this is a message").digest()

        c0 = cosi.commit(self.client, parse_path("10018'/0'"), digest)
        c1 = cosi.commit(self.client, parse_path("10018'/1'"), digest)
        c2 = cosi.commit(self.client, parse_path("10018'/2'"), digest)

        assert c0.pubkey != c1.pubkey
        assert c0.pubkey != c2.pubkey
        assert c1.pubkey != c2.pubkey

        assert c0.commitment != c1.commitment
        assert c0.commitment != c2.commitment
        assert c1.commitment != c2.commitment

        digestb = sha256(b"this is a different message").digest()

        c0b = cosi.commit(self.client, parse_path("10018'/0'"), digestb)
        c1b = cosi.commit(self.client, parse_path("10018'/1'"), digestb)
        c2b = cosi.commit(self.client, parse_path("10018'/2'"), digestb)

        assert c0.pubkey == c0b.pubkey
        assert c1.pubkey == c1b.pubkey
        assert c2.pubkey == c2b.pubkey

        assert c0.commitment != c0b.commitment
        assert c1.commitment != c1b.commitment
        assert c2.commitment != c2b.commitment

    def test_cosi_sign(self):
        self.setup_mnemonic_pin_passphrase()

        digest = sha256(b"this is a message").digest()

        c0 = cosi.commit(self.client, parse_path("10018'/0'"), digest)
        c1 = cosi.commit(self.client, parse_path("10018'/1'"), digest)
        c2 = cosi.commit(self.client, parse_path("10018'/2'"), digest)

        global_pk = cosi.combine_keys([c0.pubkey, c1.pubkey, c2.pubkey])
        global_R = cosi.combine_keys([c0.commitment, c1.commitment, c2.commitment])

        # fmt: off
        sig0 = cosi.sign(self.client, parse_path("10018'/0'"), digest, global_R, global_pk)
        sig1 = cosi.sign(self.client, parse_path("10018'/1'"), digest, global_R, global_pk)
        sig2 = cosi.sign(self.client, parse_path("10018'/2'"), digest, global_R, global_pk)
        # fmt: on

        sig = cosi.combine_sig(
            global_R, [sig0.signature, sig1.signature, sig2.signature]
        )

        cosi.verify(sig, digest, global_pk)

    def test_cosi_compat(self):
        self.setup_mnemonic_pin_passphrase()

        digest = sha256(b"this is not a pipe").digest()
        remote_commit = cosi.commit(self.client, parse_path("10018'/0'"), digest)

        local_privkey = sha256(b"private key").digest()[:32]
        local_pubkey = cosi.pubkey_from_privkey(local_privkey)
        local_nonce, local_commitment = cosi.get_nonce(local_privkey, digest, 42)

        global_pk = cosi.combine_keys([remote_commit.pubkey, local_pubkey])
        global_R = cosi.combine_keys([remote_commit.commitment, local_commitment])

        remote_sig = cosi.sign(
            self.client, parse_path("10018'/0'"), digest, global_R, global_pk
        )
        local_sig = cosi.sign_with_privkey(
            digest, local_privkey, global_pk, local_nonce, global_R
        )
        sig = cosi.combine_sig(global_R, [remote_sig.signature, local_sig])

        cosi.verify(sig, digest, global_pk)



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None  # type: ignore


class NEMMosaicDefinition(p.MessageType):

    def __init__(
        self,
        name: str = None,
        ticker: str = None,
        namespace: str = None,
        mosaic: str = None,
        divisibility: int = None,
        levy: int = None,
        fee: int = None,
        levy_address: str = None,
        levy_namespace: str = None,
        levy_mosaic: str = None,
        supply: int = None,
        mutable_supply: bool = None,
        transferable: bool = None,
        description: str = None,
        networks: List[int] = None,
    ) -> None:
        self.name = name
        self.ticker = ticker
        self.namespace = namespace
        self.mosaic = mosaic
        self.divisibility = divisibility
        self.levy = levy
        self.fee = fee
        self.levy_address = levy_address
        self.levy_namespace = levy_namespace
        self.levy_mosaic = levy_mosaic
        self.supply = supply
        self.mutable_supply = mutable_supply
        self.transferable = transferable
        self.description = description
        self.networks = networks if networks is not None else []

    @classmethod
    def get_fields(cls):
        return {
            1: ('name', p.UnicodeType, 0),
            2: ('ticker', p.UnicodeType, 0),
            3: ('namespace', p.UnicodeType, 0),
            4: ('mosaic', p.UnicodeType, 0),
            5: ('divisibility', p.UVarintType, 0),
            6: ('levy', p.UVarintType, 0),
            7: ('fee', p.UVarintType, 0),
            8: ('levy_address', p.UnicodeType, 0),
            9: ('levy_namespace', p.UnicodeType, 0),
            10: ('levy_mosaic', p.UnicodeType, 0),
            11: ('supply', p.UVarintType, 0),
            12: ('mutable_supply', p.BoolType, 0),
            13: ('transferable', p.BoolType, 0),
            14: ('description', p.UnicodeType, 0),
            15: ('networks', p.UVarintType, p.FLAG_REPEATED),
        }



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import pytest

from trezorlib import btc, debuglink

from .common import TrezorTest


@pytest.mark.skip_t2
class TestDeviceLoadXprv(TrezorTest):
    def test_load_device_xprv_1(self):
        debuglink.load_device_by_xprv(
            self.client,
            xprv="xprv9s21ZrQH143K2JF8RafpqtKiTbsbaxEeUaMnNHsm5o6wCW3z8ySyH4UxFVSfZ8n7ESu7fgir8imbZKLYVBxFPND1pniTZ81vKfd45EHKX73",
            pin="",
            passphrase_protection=False,
            label="test",
            language="english",
        )

        passphrase_protection = self.client.debug.read_passphrase_protection()
        assert passphrase_protection is False

        address = btc.get_address(self.client, "Bitcoin", [])
        assert address == "128RdrAkJDmqasgvfRf6MC5VcX4HKqH4mR"

    def test_load_device_xprv_2(self):
        debuglink.load_device_by_xprv(
            self.client,
            xprv="xprv9s21ZrQH143K2JF8RafpqtKiTbsbaxEeUaMnNHsm5o6wCW3z8ySyH4UxFVSfZ8n7ESu7fgir8imbZKLYVBxFPND1pniTZ81vKfd45EHKX73",
            pin="",
            passphrase_protection=True,
            label="test",
            language="english",
        )

        self.client.set_passphrase("passphrase")

        passphrase_protection = self.client.debug.read_passphrase_protection()
        assert passphrase_protection is True

        address = btc.get_address(self.client, "Bitcoin", [])
        assert address == "1CHUbFa4wTTPYgkYaw2LHSd5D4qJjMU8ri"



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class CardanoSignedTx(p.MessageType):
    MESSAGE_WIRE_TYPE = 310

    def __init__(
        self,
        tx_hash: bytes = None,
        tx_body: bytes = None,
    ) -> None:
        self.tx_hash = tx_hash
        self.tx_body = tx_body

    @classmethod
    def get_fields(cls):
        return {
            1: ('tx_hash', p.BytesType, 0),
            2: ('tx_body', p.BytesType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class StellarAllowTrustOp(p.MessageType):
    MESSAGE_WIRE_TYPE = 217

    def __init__(
        self,
        source_account: str = None,
        trusted_account: str = None,
        asset_type: int = None,
        asset_code: str = None,
        is_authorized: int = None,
    ) -> None:
        self.source_account = source_account
        self.trusted_account = trusted_account
        self.asset_type = asset_type
        self.asset_code = asset_code
        self.is_authorized = is_authorized

    @classmethod
    def get_fields(cls):
        return {
            1: ('source_account', p.UnicodeType, 0),
            2: ('trusted_account', p.UnicodeType, 0),
            3: ('asset_type', p.UVarintType, 0),
            4: ('asset_code', p.UnicodeType, 0),
            5: ('is_authorized', p.UVarintType, 0),
        }



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import pytest

from trezorlib import misc

from .common import TrezorTest


class TestMsgCipherkeyvalue(TrezorTest):
    def test_encrypt(self):
        self.setup_mnemonic_nopin_nopassphrase()

        # different ask values
        res = misc.encrypt_keyvalue(
            self.client,
            [0, 1, 2],
            b"test",
            b"testing message!",
            ask_on_encrypt=True,
            ask_on_decrypt=True,
        )
        assert res.hex() == "676faf8f13272af601776bc31bc14e8f"

        res = misc.encrypt_keyvalue(
            self.client,
            [0, 1, 2],
            b"test",
            b"testing message!",
            ask_on_encrypt=True,
            ask_on_decrypt=False,
        )
        assert res.hex() == "5aa0fbcb9d7fa669880745479d80c622"

        res = misc.encrypt_keyvalue(
            self.client,
            [0, 1, 2],
            b"test",
            b"testing message!",
            ask_on_encrypt=False,
            ask_on_decrypt=True,
        )
        assert res.hex() == "958d4f63269b61044aaedc900c8d6208"

        res = misc.encrypt_keyvalue(
            self.client,
            [0, 1, 2],
            b"test",
            b"testing message!",
            ask_on_encrypt=False,
            ask_on_decrypt=False,
        )
        assert res.hex() == "e0cf0eb0425947000eb546cc3994bc6c"

        # different key
        res = misc.encrypt_keyvalue(
            self.client,
            [0, 1, 2],
            b"test2",
            b"testing message!",
            ask_on_encrypt=True,
            ask_on_decrypt=True,
        )
        assert res.hex() == "de247a6aa6be77a134bb3f3f925f13af"

        # different message
        res = misc.encrypt_keyvalue(
            self.client,
            [0, 1, 2],
            b"test",
            b"testing message! it is different",
            ask_on_encrypt=True,
            ask_on_decrypt=True,
        )
        assert (
            res.hex()
            == "676faf8f13272af601776bc31bc14e8f3ae1c88536bf18f1b44f1e4c2c4a613d"
        )

        # different path
        res = misc.encrypt_keyvalue(
            self.client,
            [0, 1, 3],
            b"test",
            b"testing message!",
            ask_on_encrypt=True,
            ask_on_decrypt=True,
        )
        assert res.hex() == "b4811a9d492f5355a5186ddbfccaae7b"

    def test_decrypt(self):
        self.setup_mnemonic_nopin_nopassphrase()

        # different ask values
        res = misc.decrypt_keyvalue(
            self.client,
            [0, 1, 2],
            b"test",
            bytes.fromhex("676faf8f13272af601776bc31bc14e8f"),
            ask_on_encrypt=True,
            ask_on_decrypt=True,
        )
        assert res == b"testing message!"

        res = misc.decrypt_keyvalue(
            self.client,
            [0, 1, 2],
            b"test",
            bytes.fromhex("5aa0fbcb9d7fa669880745479d80c622"),
            ask_on_encrypt=True,
            ask_on_decrypt=False,
        )
        assert res == b"testing message!"

        res = misc.decrypt_keyvalue(
            self.client,
            [0, 1, 2],
            b"test",
            bytes.fromhex("958d4f63269b61044aaedc900c8d6208"),
            ask_on_encrypt=False,
            ask_on_decrypt=True,
        )
        assert res == b"testing message!"

        res = misc.decrypt_keyvalue(
            self.client,
            [0, 1, 2],
            b"test",
            bytes.fromhex("e0cf0eb0425947000eb546cc3994bc6c"),
            ask_on_encrypt=False,
            ask_on_decrypt=False,
        )
        assert res == b"testing message!"

        # different key
        res = misc.decrypt_keyvalue(
            self.client,
            [0, 1, 2],
            b"test2",
            bytes.fromhex("de247a6aa6be77a134bb3f3f925f13af"),
            ask_on_encrypt=True,
            ask_on_decrypt=True,
        )
        assert res == b"testing message!"

        # different message
        res = misc.decrypt_keyvalue(
            self.client,
            [0, 1, 2],
            b"test",
            bytes.fromhex(
                "676faf8f13272af601776bc31bc14e8f3ae1c88536bf18f1b44f1e4c2c4a613d"
            ),
            ask_on_encrypt=True,
            ask_on_decrypt=True,
        )
        assert res == b"testing message! it is different"

        # different path
        res = misc.decrypt_keyvalue(
            self.client,
            [0, 1, 3],
            b"test",
            bytes.fromhex("b4811a9d492f5355a5186ddbfccaae7b"),
            ask_on_encrypt=True,
            ask_on_decrypt=True,
        )
        assert res == b"testing message!"

    def test_encrypt_badlen(self):
        self.setup_mnemonic_nopin_nopassphrase()
        with pytest.raises(Exception):
            misc.encrypt_keyvalue(self.client, [0, 1, 2], b"test", b"testing")

    def test_decrypt_badlen(self):
        self.setup_mnemonic_nopin_nopassphrase()
        with pytest.raises(Exception):
            misc.decrypt_keyvalue(self.client, [0, 1, 2], b"test", b"testing")



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None  # type: ignore


class TezosProposalOp(p.MessageType):

    def __init__(
        self,
        source: bytes = None,
        period: int = None,
        proposals: List[bytes] = None,
    ) -> None:
        self.source = source
        self.period = period
        self.proposals = proposals if proposals is not None else []

    @classmethod
    def get_fields(cls):
        return {
            1: ('source', p.BytesType, 0),
            2: ('period', p.UVarintType, 0),
            4: ('proposals', p.BytesType, p.FLAG_REPEATED),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .MoneroTransactionRsigData import MoneroTransactionRsigData


class MoneroTransactionAllOutSetRequest(p.MessageType):
    MESSAGE_WIRE_TYPE = 513

    def __init__(
        self,
        rsig_data: MoneroTransactionRsigData = None,
    ) -> None:
        self.rsig_data = rsig_data

    @classmethod
    def get_fields(cls):
        return {
            1: ('rsig_data', MoneroTransactionRsigData, 0),
        }



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import pytest
from mnemonic import Mnemonic

from trezorlib import device, messages as proto

from .common import TrezorTest, generate_entropy


@pytest.mark.skip_t2
class TestMsgResetDevice(TrezorTest):
    def test_reset_device(self):

        # No PIN, no passphrase
        external_entropy = b"zlutoucky kun upel divoke ody" * 2
        strength = 128

        ret = self.client.call_raw(
            proto.ResetDevice(
                display_random=False,
                strength=strength,
                passphrase_protection=False,
                pin_protection=False,
                language="english",
                label="test",
            )
        )

        assert isinstance(ret, proto.ButtonRequest)
        self.client.debug.press_yes()
        ret = self.client.call_raw(proto.ButtonAck())

        # Provide entropy
        assert isinstance(ret, proto.EntropyRequest)
        internal_entropy = self.client.debug.read_reset_entropy()
        ret = self.client.call_raw(proto.EntropyAck(entropy=external_entropy))

        # Generate mnemonic locally
        entropy = generate_entropy(strength, internal_entropy, external_entropy)
        expected_mnemonic = Mnemonic("english").to_mnemonic(entropy)

        mnemonic = []
        for _ in range(strength // 32 * 3):
            assert isinstance(ret, proto.ButtonRequest)
            mnemonic.append(self.client.debug.read_reset_word())
            self.client.debug.press_yes()
            self.client.call_raw(proto.ButtonAck())

        mnemonic = " ".join(mnemonic)

        # Compare that device generated proper mnemonic for given entropies
        assert mnemonic == expected_mnemonic

        mnemonic = []
        for _ in range(strength // 32 * 3):
            assert isinstance(ret, proto.ButtonRequest)
            mnemonic.append(self.client.debug.read_reset_word())
            self.client.debug.press_yes()
            resp = self.client.call_raw(proto.ButtonAck())

        assert isinstance(resp, proto.Success)

        mnemonic = " ".join(mnemonic)

        # Compare that second pass printed out the same mnemonic once again
        assert mnemonic == expected_mnemonic

        # Check if device is properly initialized
        resp = self.client.call_raw(proto.Initialize())
        assert resp.initialized is True
        assert resp.needs_backup is False
        assert resp.pin_protection is False
        assert resp.passphrase_protection is False

        # Do passphrase-protected action, PassphraseRequest should NOT be raised
        resp = self.client.call_raw(proto.Ping(passphrase_protection=True))
        assert isinstance(resp, proto.Success)

        # Do PIN-protected action, PinRequest should NOT be raised
        resp = self.client.call_raw(proto.Ping(pin_protection=True))
        assert isinstance(resp, proto.Success)

    def test_reset_device_pin(self):
        external_entropy = b"zlutoucky kun upel divoke ody" * 2
        strength = 128

        ret = self.client.call_raw(
            proto.ResetDevice(
                display_random=True,
                strength=strength,
                passphrase_protection=True,
                pin_protection=True,
                language="english",
                label="test",
            )
        )

        assert isinstance(ret, proto.ButtonRequest)
        self.client.debug.press_yes()
        ret = self.client.call_raw(proto.ButtonAck())

        assert isinstance(ret, proto.ButtonRequest)
        self.client.debug.press_yes()
        ret = self.client.call_raw(proto.ButtonAck())

        assert isinstance(ret, proto.PinMatrixRequest)

        # Enter PIN for first time
        pin_encoded = self.client.debug.encode_pin("654")
        ret = self.client.call_raw(proto.PinMatrixAck(pin=pin_encoded))
        assert isinstance(ret, proto.PinMatrixRequest)

        # Enter PIN for second time
        pin_encoded = self.client.debug.encode_pin("654")
        ret = self.client.call_raw(proto.PinMatrixAck(pin=pin_encoded))

        # Provide entropy
        assert isinstance(ret, proto.EntropyRequest)
        internal_entropy = self.client.debug.read_reset_entropy()
        ret = self.client.call_raw(proto.EntropyAck(entropy=external_entropy))

        # Generate mnemonic locally
        entropy = generate_entropy(strength, internal_entropy, external_entropy)
        expected_mnemonic = Mnemonic("english").to_mnemonic(entropy)

        mnemonic = []
        for _ in range(strength // 32 * 3):
            assert isinstance(ret, proto.ButtonRequest)
            mnemonic.append(self.client.debug.read_reset_word())
            self.client.debug.press_yes()
            self.client.call_raw(proto.ButtonAck())

        mnemonic = " ".join(mnemonic)

        # Compare that device generated proper mnemonic for given entropies
        assert mnemonic == expected_mnemonic

        mnemonic = []
        for _ in range(strength // 32 * 3):
            assert isinstance(ret, proto.ButtonRequest)
            mnemonic.append(self.client.debug.read_reset_word())
            self.client.debug.press_yes()
            resp = self.client.call_raw(proto.ButtonAck())

        assert isinstance(resp, proto.Success)

        mnemonic = " ".join(mnemonic)

        # Compare that second pass printed out the same mnemonic once again
        assert mnemonic == expected_mnemonic

        # Check if device is properly initialized
        resp = self.client.call_raw(proto.Initialize())
        assert resp.initialized is True
        assert resp.needs_backup is False
        assert resp.pin_protection is True
        assert resp.passphrase_protection is True

        # Do passphrase-protected action, PassphraseRequest should be raised
        resp = self.client.call_raw(proto.Ping(passphrase_protection=True))
        assert isinstance(resp, proto.PassphraseRequest)
        self.client.call_raw(proto.Cancel())

    def test_failed_pin(self):
        # external_entropy = b'zlutoucky kun upel divoke ody' * 2
        strength = 128

        ret = self.client.call_raw(
            proto.ResetDevice(
                display_random=True,
                strength=strength,
                passphrase_protection=True,
                pin_protection=True,
                language="english",
                label="test",
            )
        )

        assert isinstance(ret, proto.ButtonRequest)
        self.client.debug.press_yes()
        ret = self.client.call_raw(proto.ButtonAck())

        assert isinstance(ret, proto.ButtonRequest)
        self.client.debug.press_yes()
        ret = self.client.call_raw(proto.ButtonAck())

        assert isinstance(ret, proto.PinMatrixRequest)

        # Enter PIN for first time
        pin_encoded = self.client.debug.encode_pin(self.pin4)
        ret = self.client.call_raw(proto.PinMatrixAck(pin=pin_encoded))
        assert isinstance(ret, proto.PinMatrixRequest)

        # Enter PIN for second time
        pin_encoded = self.client.debug.encode_pin(self.pin6)
        ret = self.client.call_raw(proto.PinMatrixAck(pin=pin_encoded))

        assert isinstance(ret, proto.Failure)

    def test_already_initialized(self):
        self.setup_mnemonic_nopin_nopassphrase()
        with pytest.raises(Exception):
            device.reset(self.client, False, 128, True, True, "label", "english")



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class MoneroLiveRefreshStepAck(p.MessageType):
    MESSAGE_WIRE_TYPE = 555

    def __init__(
        self,
        salt: bytes = None,
        key_image: bytes = None,
    ) -> None:
        self.salt = salt
        self.key_image = key_image

    @classmethod
    def get_fields(cls):
        return {
            1: ('salt', p.BytesType, 0),
            2: ('key_image', p.BytesType, 0),
        }



# Automatically generated by pb2py
# fmt: off
UnexpectedMessage = 1
ButtonExpected = 2
DataError = 3
ActionCancelled = 4
PinExpected = 5
PinCancelled = 6
PinInvalid = 7
InvalidSignature = 8
ProcessError = 9
NotEnoughFunds = 10
NotInitialized = 11
PinMismatch = 12
FirmwareError = 99



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class Ping(p.MessageType):
    MESSAGE_WIRE_TYPE = 1

    def __init__(
        self,
        message: str = None,
        button_protection: bool = None,
        pin_protection: bool = None,
        passphrase_protection: bool = None,
    ) -> None:
        self.message = message
        self.button_protection = button_protection
        self.pin_protection = pin_protection
        self.passphrase_protection = passphrase_protection

    @classmethod
    def get_fields(cls):
        return {
            1: ('message', p.UnicodeType, 0),
            2: ('button_protection', p.BoolType, 0),
            3: ('pin_protection', p.BoolType, 0),
            4: ('passphrase_protection', p.BoolType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class EosPublicKey(p.MessageType):
    MESSAGE_WIRE_TYPE = 601

    def __init__(
        self,
        wif_public_key: str = None,
        raw_public_key: bytes = None,
    ) -> None:
        self.wif_public_key = wif_public_key
        self.raw_public_key = raw_public_key

    @classmethod
    def get_fields(cls):
        return {
            1: ('wif_public_key', p.UnicodeType, 0),
            2: ('raw_public_key', p.BytesType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class EosActionBuyRamBytes(p.MessageType):

    def __init__(
        self,
        payer: int = None,
        receiver: int = None,
        bytes: int = None,
    ) -> None:
        self.payer = payer
        self.receiver = receiver
        self.bytes = bytes

    @classmethod
    def get_fields(cls):
        return {
            1: ('payer', p.UVarintType, 0),
            2: ('receiver', p.UVarintType, 0),
            3: ('bytes', p.UVarintType, 0),
        }



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import pytest

from trezorlib.cardano import get_address
from trezorlib.tools import parse_path

from .common import TrezorTest


@pytest.mark.cardano
@pytest.mark.skip_t1  # T1 support is not planned
class TestMsgCardanoGetAddress(TrezorTest):
    @pytest.mark.parametrize(
        "path,expected_address",
        [
            (
                "m/44'/1815'/0'/0/0",
                "Ae2tdPwUPEZLCq3sFv4wVYxwqjMH2nUzBVt1HFr4v87snYrtYq3d3bq2PUQ",
            ),
            (
                "m/44'/1815'/0'/0/1",
                "Ae2tdPwUPEZEY6pVJoyuNNdLp7VbMB7U7qfebeJ7XGunk5Z2eHarkcN1bHK",
            ),
            (
                "m/44'/1815'/0'/0/2",
                "Ae2tdPwUPEZ3gZD1QeUHvAqadAV59Zid6NP9VCR9BG5LLAja9YtBUgr6ttK",
            ),
        ],
    )
    def test_cardano_get_address(self, path, expected_address):
        # data from https://iancoleman.io/bip39/#english
        self.setup_mnemonic_nopin_nopassphrase()

        address = get_address(self.client, parse_path(path))
        assert address == expected_address



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class DebugLinkLog(p.MessageType):
    MESSAGE_WIRE_TYPE = 104

    def __init__(
        self,
        level: int = None,
        bucket: str = None,
        text: str = None,
    ) -> None:
        self.level = level
        self.bucket = bucket
        self.text = text

    @classmethod
    def get_fields(cls):
        return {
            1: ('level', p.UVarintType, 0),
            2: ('bucket', p.UnicodeType, 0),
            3: ('text', p.UnicodeType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class EosActionUnknown(p.MessageType):

    def __init__(
        self,
        data_size: int = None,
        data_chunk: bytes = None,
    ) -> None:
        self.data_size = data_size
        self.data_chunk = data_chunk

    @classmethod
    def get_fields(cls):
        return {
            1: ('data_size', p.UVarintType, 0),
            2: ('data_chunk', p.BytesType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class MoneroTransactionSetInputAck(p.MessageType):
    MESSAGE_WIRE_TYPE = 504

    def __init__(
        self,
        vini: bytes = None,
        vini_hmac: bytes = None,
        pseudo_out: bytes = None,
        pseudo_out_hmac: bytes = None,
        pseudo_out_alpha: bytes = None,
        spend_key: bytes = None,
    ) -> None:
        self.vini = vini
        self.vini_hmac = vini_hmac
        self.pseudo_out = pseudo_out
        self.pseudo_out_hmac = pseudo_out_hmac
        self.pseudo_out_alpha = pseudo_out_alpha
        self.spend_key = spend_key

    @classmethod
    def get_fields(cls):
        return {
            1: ('vini', p.BytesType, 0),
            2: ('vini_hmac', p.BytesType, 0),
            3: ('pseudo_out', p.BytesType, 0),
            4: ('pseudo_out_hmac', p.BytesType, 0),
            5: ('pseudo_out_alpha', p.BytesType, 0),
            6: ('spend_key', p.BytesType, 0),
        }



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

from . import messages as proto
from .protobuf import dict_to_proto
from .tools import CallException, dict_from_camelcase, expect, normalize_nfc


@expect(proto.LiskAddress, field="address")
def get_address(client, n, show_display=False):
    return client.call(proto.LiskGetAddress(address_n=n, show_display=show_display))


@expect(proto.LiskPublicKey)
def get_public_key(client, n, show_display=False):
    return client.call(proto.LiskGetPublicKey(address_n=n, show_display=show_display))


@expect(proto.LiskMessageSignature)
def sign_message(client, n, message):
    message = normalize_nfc(message)
    return client.call(proto.LiskSignMessage(address_n=n, message=message))


def verify_message(client, pubkey, signature, message):
    message = normalize_nfc(message)
    try:
        resp = client.call(
            proto.LiskVerifyMessage(
                signature=signature, public_key=pubkey, message=message
            )
        )
    except CallException as e:
        resp = e
    return isinstance(resp, proto.Success)


RENAMES = {"lifetime": "life_time", "keysgroup": "keys_group"}


@expect(proto.LiskSignedTx)
def sign_tx(client, n, transaction):
    transaction = dict_from_camelcase(transaction, renames=RENAMES)
    msg = dict_to_proto(proto.LiskTransactionCommon, transaction)
    return client.call(proto.LiskSignTx(address_n=n, transaction=msg))



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class LiskDelegateType(p.MessageType):

    def __init__(
        self,
        username: str = None,
    ) -> None:
        self.username = username

    @classmethod
    def get_fields(cls):
        return {
            1: ('username', p.UnicodeType, 0),
        }



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import base64

from trezorlib import messages, stellar


def test_stellar_parse_transaction_bytes_simple():
    b64 = b"AAAAABXWSL/k028ZbPtXNf/YylTNS4Iz90PyJEnefPMBzbRpAAAAZAAAAAEAAAAAAAAAAAAAAAAAAAABAAAAAAAAAAAAAAAAXVVkJGaxhbhDFS6eIZFR28WJICfsQBAaUXvtXKAwwuAAAAAAO5/eyAAAAAA="

    tx, operations = stellar.parse_transaction_bytes(base64.b64decode(b64))

    assert (
        tx.source_account == "GAK5MSF74TJW6GLM7NLTL76YZJKM2S4CGP3UH4REJHPHZ4YBZW2GSBPW"
    )
    assert tx.fee == 100
    assert tx.sequence_number == 4294967296
    assert tx.timebounds_start is None
    assert tx.timebounds_end is None
    assert tx.memo_type == stellar.MEMO_TYPE_NONE
    assert tx.memo_text is None
    assert tx.memo_id is None
    assert tx.memo_hash is None
    assert tx.num_operations == len(operations)


def test_stellar_parse_transaction_bytes_memo_text():
    b64 = b"AAAAABXWSL/k028ZbPtXNf/YylTNS4Iz90PyJEnefPMBzbRpAAAAZAAAAAEAAAAAAAAAAAAAAAEAAAAMZXhhbXBsZSBtZW1vAAAAAQAAAAAAAAAAAAAAAF1VZCRmsYW4QxUuniGRUdvFiSAn7EAQGlF77VygMMLgAAAAADuf3sgAAAAA"

    tx, operations = stellar.parse_transaction_bytes(base64.b64decode(b64))

    assert (
        tx.source_account == "GAK5MSF74TJW6GLM7NLTL76YZJKM2S4CGP3UH4REJHPHZ4YBZW2GSBPW"
    )
    assert tx.fee == 100
    assert tx.sequence_number == 4294967296
    assert tx.timebounds_start is None
    assert tx.timebounds_end is None
    assert tx.memo_type == stellar.MEMO_TYPE_TEXT
    assert tx.memo_text == b"example memo"
    assert tx.memo_id is None
    assert tx.memo_hash is None
    assert tx.num_operations == len(operations)


def test_stellar_parse_transaction_bytes_memo_id():
    b64 = b"AAAAABXWSL/k028ZbPtXNf/YylTNS4Iz90PyJEnefPMBzbRpAAAAZAAAAAEAAAAAAAAAAAAAAAIAAAAAB1vNFQAAAAEAAAAAAAAAAAAAAABdVWQkZrGFuEMVLp4hkVHbxYkgJ+xAEBpRe+1coDDC4AAAAAA7n97IAAAAAA=="

    tx, operations = stellar.parse_transaction_bytes(base64.b64decode(b64))

    assert (
        tx.source_account == "GAK5MSF74TJW6GLM7NLTL76YZJKM2S4CGP3UH4REJHPHZ4YBZW2GSBPW"
    )
    assert tx.fee == 100
    assert tx.sequence_number == 4294967296
    assert tx.timebounds_start is None
    assert tx.timebounds_end is None
    assert tx.memo_type == stellar.MEMO_TYPE_ID
    assert tx.memo_text is None
    assert tx.memo_id == 123456789
    assert tx.memo_hash is None
    assert tx.num_operations == len(operations)


def test_stellar_parse_transaction_bytes_memo_hash():
    b64 = b"AAAAABXWSL/k028ZbPtXNf/YylTNS4Iz90PyJEnefPMBzbRpAAAAZAAAAAEAAAAAAAAAAAAAAAMjLtb5+r8U47tVOSsYz+PQ/ryU0gzGMnw4odB11uoRjAAAAAEAAAAAAAAAAAAAAABdVWQkZrGFuEMVLp4hkVHbxYkgJ+xAEBpRe+1coDDC4AAAAAA7n97IAAAAAA=="

    tx, operations = stellar.parse_transaction_bytes(base64.b64decode(b64))

    assert (
        tx.source_account == "GAK5MSF74TJW6GLM7NLTL76YZJKM2S4CGP3UH4REJHPHZ4YBZW2GSBPW"
    )
    assert tx.fee == 100
    assert tx.sequence_number == 4294967296
    assert tx.timebounds_start is None
    assert tx.timebounds_end is None
    assert tx.memo_type == stellar.MEMO_TYPE_HASH
    assert tx.memo_text is None
    assert tx.memo_id is None
    # base-64 encoding of the raw bytes of sha256('stellar')
    assert (
        base64.b64encode(tx.memo_hash)
        == b"Iy7W+fq/FOO7VTkrGM/j0P68lNIMxjJ8OKHQddbqEYw="
    )
    assert tx.num_operations == len(operations)


def test_stellar_parse_transaction_bytes_memo_return():
    b64 = b"AAAAABXWSL/k028ZbPtXNf/YylTNS4Iz90PyJEnefPMBzbRpAAAAZAAAAAEAAAAAAAAAAAAAAAQjLtb5+r8U47tVOSsYz+PQ/ryU0gzGMnw4odB11uoRjAAAAAEAAAAAAAAAAAAAAABdVWQkZrGFuEMVLp4hkVHbxYkgJ+xAEBpRe+1coDDC4AAAAAA7n97IAAAAAA=="

    tx, operations = stellar.parse_transaction_bytes(base64.b64decode(b64))

    assert (
        tx.source_account == "GAK5MSF74TJW6GLM7NLTL76YZJKM2S4CGP3UH4REJHPHZ4YBZW2GSBPW"
    )
    assert tx.fee == 100
    assert tx.sequence_number == 4294967296
    assert tx.timebounds_start is None
    assert tx.timebounds_end is None
    assert tx.memo_type == stellar.MEMO_TYPE_RETURN
    assert tx.memo_text is None
    assert tx.memo_id is None
    # base-64 encoding of the raw bytes of sha256('stellar')
    assert (
        base64.b64encode(tx.memo_hash)
        == b"Iy7W+fq/FOO7VTkrGM/j0P68lNIMxjJ8OKHQddbqEYw="
    )
    assert tx.num_operations == len(operations)


def test_stellar_parse_operation_bytes_create_account_simple():
    b64 = b"AAAAABXWSL/k028ZbPtXNf/YylTNS4Iz90PyJEnefPMBzbRpAAAAZAAAAAEAAAAAAAAAAAAAAAAAAAABAAAAAAAAAAAAAAAAXVVkJGaxhbhDFS6eIZFR28WJICfsQBAaUXvtXKAwwuAAAAAAO5/eyAAAAAA="

    tx, operations = stellar.parse_transaction_bytes(base64.b64decode(b64))
    op = operations[0]

    assert isinstance(op, messages.StellarCreateAccountOp)
    assert op.source_account is None
    assert op.new_account == "GBOVKZBEM2YYLOCDCUXJ4IMRKHN4LCJAE7WEAEA2KF562XFAGDBOB64V"
    assert op.starting_balance == 1000333000


def test_stellar_parse_operation_bytes_payment_native():
    b64 = b"AAAAABXWSL/k028ZbPtXNf/YylTNS4Iz90PyJEnefPMBzbRpAAAAZAAAAAEAAAAAAAAAAAAAAAAAAAABAAAAAAAAAAEAAAAAXVVkJGaxhbhDFS6eIZFR28WJICfsQBAaUXvtXKAwwuAAAAAAAAAAAB3PFpgAAAAA"

    tx, operations = stellar.parse_transaction_bytes(base64.b64decode(b64))
    op = operations[0]

    assert isinstance(op, messages.StellarPaymentOp)
    assert op.source_account is None
    assert (
        op.destination_account
        == "GBOVKZBEM2YYLOCDCUXJ4IMRKHN4LCJAE7WEAEA2KF562XFAGDBOB64V"
    )
    assert op.asset.type == stellar.ASSET_TYPE_NATIVE
    assert op.amount == 500111000


def test_stellar_parse_operation_bytes_payment_custom4():
    b64 = b"AAAAABXWSL/k028ZbPtXNf/YylTNS4Iz90PyJEnefPMBzbRpAAAAZAAAAAEAAAAAAAAAAAAAAAAAAAABAAAAAAAAAAEAAAAAXVVkJGaxhbhDFS6eIZFR28WJICfsQBAaUXvtXKAwwuAAAAABVEVTVAAAAAAphJYCwg5YNl8SPBLYehykVQ0QzSGwrg4Y1E4+Vv1qFQAAAAAdzxaYAAAAAA=="

    tx, operations = stellar.parse_transaction_bytes(base64.b64decode(b64))
    op = operations[0]

    assert op.source_account is None
    assert (
        op.destination_account
        == "GBOVKZBEM2YYLOCDCUXJ4IMRKHN4LCJAE7WEAEA2KF562XFAGDBOB64V"
    )
    assert op.asset.type == stellar.ASSET_TYPE_ALPHA4
    assert op.asset.code == b"TEST"
    assert op.asset.issuer == "GAUYJFQCYIHFQNS7CI6BFWD2DSSFKDIQZUQ3BLQODDKE4PSW7VVBKENC"
    assert op.amount == 500111000


def test_stellar_parse_operation_bytes_payment_custom7():
    b64 = b"AAAAABXWSL/k028ZbPtXNf/YylTNS4Iz90PyJEnefPMBzbRpAAAAZAAAAAEAAAAAAAAAAAAAAAAAAAABAAAAAAAAAAEAAAAAXVVkJGaxhbhDFS6eIZFR28WJICfsQBAaUXvtXKAwwuAAAAACU0VWRU5YWAAAAAAAAAAAACmElgLCDlg2XxI8Eth6HKRVDRDNIbCuDhjUTj5W/WoVAAAAAB3PFpgAAAAA"

    tx, operations = stellar.parse_transaction_bytes(base64.b64decode(b64))
    op = operations[0]

    assert isinstance(op, messages.StellarPaymentOp)
    assert op.source_account is None
    assert (
        op.destination_account
        == "GBOVKZBEM2YYLOCDCUXJ4IMRKHN4LCJAE7WEAEA2KF562XFAGDBOB64V"
    )
    assert op.asset.type == stellar.ASSET_TYPE_ALPHA12
    # asset codes are either 4 or 12 characters, so this will be null-padded at the end
    assert op.asset.code == b"SEVENXX\x00\x00\x00\x00\x00"
    assert op.asset.issuer == "GAUYJFQCYIHFQNS7CI6BFWD2DSSFKDIQZUQ3BLQODDKE4PSW7VVBKENC"
    assert op.amount == 500111000


def test_stellar_parse_operation_bytes_path_payment_none():
    b64 = b"AAAAABXWSL/k028ZbPtXNf/YylTNS4Iz90PyJEnefPMBzbRpAAAAZAAAAAEAAAAAAAAAAAAAAAAAAAABAAAAAAAAAAIAAAAAAAAAAHfOKn8AAAAAXVVkJGaxhbhDFS6eIZFR28WJICfsQBAaUXvtXKAwwuAAAAABSlBZAAAAAADE+xa3Eb3cy85WSdqgwnUtC6UDwrC41YDANuCqe8vGxgAAAAAL68IBAAAAAAAAAAA="

    tx, operations = stellar.parse_transaction_bytes(base64.b64decode(b64))
    op = operations[0]

    assert isinstance(op, messages.StellarPathPaymentOp)
    assert op.source_account is None
    assert (
        op.destination_account
        == "GBOVKZBEM2YYLOCDCUXJ4IMRKHN4LCJAE7WEAEA2KF562XFAGDBOB64V"
    )

    assert op.send_asset.type == stellar.ASSET_TYPE_NATIVE
    assert op.send_max == 2009999999

    assert (
        op.destination_account
        == "GBOVKZBEM2YYLOCDCUXJ4IMRKHN4LCJAE7WEAEA2KF562XFAGDBOB64V"
    )
    assert op.destination_asset.type == stellar.ASSET_TYPE_ALPHA4
    # asset codes are either 4 or 12 characters, so this will be null-padded at the end
    assert op.destination_asset.code == b"JPY\x00"
    assert (
        op.destination_asset.issuer
        == "GDCPWFVXCG65ZS6OKZE5VIGCOUWQXJIDYKYLRVMAYA3OBKT3ZPDMNTIJ"
    )

    assert len(op.paths) == 0


def test_stellar_parse_operation_bytes_path_payment_one():
    b64 = b"AAAAABXWSL/k028ZbPtXNf/YylTNS4Iz90PyJEnefPMBzbRpAAAAZAAAAAEAAAAAAAAAAAAAAAAAAAABAAAAAAAAAAIAAAAAAAAAAHfOKn8AAAAAXVVkJGaxhbhDFS6eIZFR28WJICfsQBAaUXvtXKAwwuAAAAABSlBZAAAAAADE+xa3Eb3cy85WSdqgwnUtC6UDwrC41YDANuCqe8vGxgAAAAAL68IBAAAAAQAAAAFQVEgxAAAAAMz/d9fJ3rFifblw3jT7sRZv/Ja+fqLfob//aLZQRQibAAAAAA=="

    tx, operations = stellar.parse_transaction_bytes(base64.b64decode(b64))
    op = operations[0]

    assert isinstance(op, messages.StellarPathPaymentOp)
    assert op.source_account is None
    assert (
        op.destination_account
        == "GBOVKZBEM2YYLOCDCUXJ4IMRKHN4LCJAE7WEAEA2KF562XFAGDBOB64V"
    )

    assert op.send_asset.type == stellar.ASSET_TYPE_NATIVE
    assert op.send_max == 2009999999

    assert (
        op.destination_account
        == "GBOVKZBEM2YYLOCDCUXJ4IMRKHN4LCJAE7WEAEA2KF562XFAGDBOB64V"
    )
    assert op.destination_asset.type == stellar.ASSET_TYPE_ALPHA4
    # asset codes are either 4 or 12 characters, so this will be null-padded at the end
    assert op.destination_asset.code == b"JPY\x00"
    assert (
        op.destination_asset.issuer
        == "GDCPWFVXCG65ZS6OKZE5VIGCOUWQXJIDYKYLRVMAYA3OBKT3ZPDMNTIJ"
    )
    assert op.destination_amount == 200000001

    assert len(op.paths) == 1
    assert op.paths[0].type == stellar.ASSET_TYPE_ALPHA4
    assert op.paths[0].code == b"PTH1"
    assert (
        op.paths[0].issuer == "GDGP656XZHPLCYT5XFYN4NH3WELG77EWXZ7KFX5BX77WRNSQIUEJXAJK"
    )


def test_stellar_parse_operation_bytes_manage_offer_new():
    b64 = b"AAAAABXWSL/k028ZbPtXNf/YylTNS4Iz90PyJEnefPMBzbRpAAAAZAAAAAEAAAAAAAAAAAAAAAAAAAABAAAAAAAAAAMAAAAAAAAAAVVTRAAAAAAABkAD8fq0d+bofA1LCatUL0dCTJexnyYYd4Y1ghnNUXMAAAAAdzWUAAAKSzYAD0JAAAAAAAAAAAAAAAAA"

    tx, operations = stellar.parse_transaction_bytes(base64.b64decode(b64))
    op = operations[0]

    assert isinstance(op, messages.StellarManageOfferOp)
    assert op.source_account is None

    assert op.selling_asset.type == stellar.ASSET_TYPE_NATIVE

    assert op.buying_asset.type == stellar.ASSET_TYPE_ALPHA4
    # asset codes are either 4 or 12 characters, so this will be null-padded at the end
    assert op.buying_asset.code == b"USD\x00"
    assert (
        op.buying_asset.issuer
        == "GADEAA7R7K2HPZXIPQGUWCNLKQXUOQSMS6YZ6JQYO6DDLAQZZVIXG74A"
    )

    assert op.amount == 2000000000
    assert op.price_n == 674614
    assert op.price_d == 1000000
    assert op.offer_id == 0  # indicates a new offer


def test_stellar_parse_operation_bytes_passive_offer_new():
    b64 = b"AAAAABXWSL/k028ZbPtXNf/YylTNS4Iz90PyJEnefPMBzbRpAAAAZAAAAAEAAAAAAAAAAAAAAAAAAAABAAAAAAAAAAQAAAAAAAAAAVVTRAAAAAAABkAD8fq0d+bofA1LCatUL0dCTJexnyYYd4Y1ghnNUXMAAAAAdzWUAAAKSzYAD0JAAAAAAA=="

    tx, operations = stellar.parse_transaction_bytes(base64.b64decode(b64))
    op = operations[0]

    assert isinstance(op, messages.StellarCreatePassiveOfferOp)
    assert op.source_account is None

    assert op.selling_asset.type == stellar.ASSET_TYPE_NATIVE

    assert op.buying_asset.type == stellar.ASSET_TYPE_ALPHA4
    # asset codes are either 4 or 12 characters, so this will be null-padded at the end
    assert op.buying_asset.code == b"USD\x00"
    assert (
        op.buying_asset.issuer
        == "GADEAA7R7K2HPZXIPQGUWCNLKQXUOQSMS6YZ6JQYO6DDLAQZZVIXG74A"
    )

    assert op.amount == 2000000000
    assert op.price_n == 674614
    assert op.price_d == 1000000


def test_stellar_parse_operation_bytes_set_options_inflation():
    b64 = b"AAAAABXWSL/k028ZbPtXNf/YylTNS4Iz90PyJEnefPMBzbRpAAAAZAAAAAEAAAAAAAAAAAAAAAAAAAABAAAAAAAAAAUAAAABAAAAAAt5i66vbwH70/2M4Oj0rQW81SNLAjfOsMV2bavzocXhAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"

    tx, operations = stellar.parse_transaction_bytes(base64.b64decode(b64))
    op = operations[0]

    assert isinstance(op, messages.StellarSetOptionsOp)
    assert op.source_account is None

    assert (
        op.inflation_destination_account
        == "GAFXTC5OV5XQD66T7WGOB2HUVUC3ZVJDJMBDPTVQYV3G3K7TUHC6CLBR"
    )


def test_stellar_parse_operation_bytes_change_trust_add():
    b64 = b"AAAAABXWSL/k028ZbPtXNf/YylTNS4Iz90PyJEnefPMBzbRpAAAAZAAAAAEAAAAAAAAAAAAAAAAAAAABAAAAAAAAAAYAAAABVVNEAAAAAACkn7CoQZEWAlyO6z6VBUAddrDDR078TtLt/nP/hZJ9KQAAAAJUC+QAAAAAAA=="

    tx, operations = stellar.parse_transaction_bytes(base64.b64decode(b64))
    op = operations[0]

    assert isinstance(op, messages.StellarChangeTrustOp)
    assert op.source_account is None

    assert op.asset.type == stellar.ASSET_TYPE_ALPHA4
    assert op.asset.code == b"USD\x00"
    assert op.asset.issuer == "GCSJ7MFIIGIRMAS4R3VT5FIFIAOXNMGDI5HPYTWS5X7HH74FSJ6STSGF"

    assert op.limit == 10000000000


def test_stellar_parse_operation_bytes_allow_trust_allow():
    b64 = b"AAAAABXWSL/k028ZbPtXNf/YylTNS4Iz90PyJEnefPMBzbRpAAAAZAAAAAEAAAAAAAAAAAAAAAAAAAABAAAAAAAAAAcAAAAAZ0Me3OnxI2tuaC8qt95THF1fuB42qARTnP2ookJapQUAAAABSlBZAAAAAAEAAAAA"

    tx, operations = stellar.parse_transaction_bytes(base64.b64decode(b64))
    op = operations[0]

    assert isinstance(op, messages.StellarAllowTrustOp)
    assert op.source_account is None

    assert op.asset_type == stellar.ASSET_TYPE_ALPHA4
    assert op.asset_code == b"JPY\x00"

    assert (
        op.trusted_account == "GBTUGHW45HYSG23ONAXSVN66KMOF2X5YDY3KQBCTTT62RISCLKSQLYF4"
    )


def test_stellar_parse_operation_bytes_account_merge_simple():
    b64 = b"AAAAABXWSL/k028ZbPtXNf/YylTNS4Iz90PyJEnefPMBzbRpAAAAZAAAAAEAAAAAAAAAAAAAAAAAAAABAAAAAAAAAAgAAAAAXVVkJGaxhbhDFS6eIZFR28WJICfsQBAaUXvtXKAwwuAAAAAA"

    tx, operations = stellar.parse_transaction_bytes(base64.b64decode(b64))
    op = operations[0]

    assert isinstance(op, messages.StellarAccountMergeOp)
    assert op.source_account is None

    assert (
        op.destination_account
        == "GBOVKZBEM2YYLOCDCUXJ4IMRKHN4LCJAE7WEAEA2KF562XFAGDBOB64V"
    )


def test_stellar_parse_operation_bytes_manage_data_set_simple():
    b64 = b"AAAAABXWSL/k028ZbPtXNf/YylTNS4Iz90PyJEnefPMBzbRpAAAAZAAAAAEAAAAAAAAAAAAAAAAAAAABAAAAAAAAAAoAAAAJdGVzdCBkYXRhAAAAAAAAAQAAAARhc2RmAAAAAA=="

    tx, operations = stellar.parse_transaction_bytes(base64.b64decode(b64))
    op = operations[0]

    assert isinstance(op, messages.StellarManageDataOp)
    assert op.source_account is None

    assert op.key == b"test data"
    assert op.value == b"asdf"


def test_stellar_parse_operation_bytes_bump_sequence_simple():
    b64 = b"AAAAABXWSL/k028ZbPtXNf/YylTNS4Iz90PyJEnefPMBzbRpAAAAZAAAAAEAAAAAAAAAAAAAAAAAAAABAAAAAAAAAAsAAAAASZYC0gAAAAA="

    tx, operations = stellar.parse_transaction_bytes(base64.b64decode(b64))
    op = operations[0]

    assert isinstance(op, messages.StellarBumpSequenceOp)
    assert op.source_account is None

    assert op.bump_to == 1234567890



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import time

import pytest

from trezorlib import messages as proto
from trezorlib.exceptions import PinException

from .common import TrezorTest

# FIXME TODO Add passphrase tests


@pytest.mark.skip_t2
class TestProtectCall(TrezorTest):
    def _some_protected_call(self, button, pin, passphrase):
        # This method perform any call which have protection in the device
        res = self.client.ping(
            "random data",
            button_protection=button,
            pin_protection=pin,
            passphrase_protection=passphrase,
        )
        assert res == "random data"

    """
    def test_expected_responses(self):
        self.setup_mnemonic_pin_passphrase()

        # This is low-level test of set_expected_responses()
        # feature of debugging client

        with self.client:
            # Scenario 1 - Received unexpected message
            self.client.set_expected_responses([])
            with pytest.raises(CallException):
                self._some_protected_call(True, True, True)

        with self.client:
            # Scenario 2 - Received other than expected message
            self.client.set_expected_responses([proto.Success()])
            with pytest.raises(CallException):
                self._some_protected_call(True, True, True)

        def scenario3():
            with self.client:
                # Scenario 3 - Not received expected message
                self.client.set_expected_responses([proto.ButtonRequest(),
                                                    proto.Success(),
                                                    proto.Success()])  # This is expected, but not received
                self._some_protected_call(True, False, False)
                with pytest.raises(Exception):
                    scenario3()

        with self.client:
            # Scenario 4 - Received what expected
            self.client.set_expected_responses([proto.ButtonRequest(),
                                                proto.PinMatrixRequest(),
                                                proto.PassphraseRequest(),
                                                proto.Success(message='random data')])
            self._some_protected_call(True, True, True)

        def scenario5():
            with self.client:
                # Scenario 5 - Failed message by field filter
                self.client.set_expected_responses([proto.ButtonRequest(),
                                                    proto.Success(message='wrong data')])
                self._some_protected_call(True, True, True)
        with pytest.raises(CallException):
            scenario5()
    """

    def test_no_protection(self):
        self.setup_mnemonic_nopin_nopassphrase()

        with self.client:
            assert self.client.debug.read_pin()[0] is None
            self.client.set_expected_responses([proto.Success()])
            self._some_protected_call(False, True, True)

    def test_pin(self):
        self.setup_mnemonic_pin_passphrase()

        with self.client:
            assert self.client.debug.read_pin()[0] == self.pin4
            self.client.setup_debuglink(button=True, pin_correct=True)
            self.client.set_expected_responses(
                [proto.ButtonRequest(), proto.PinMatrixRequest(), proto.Success()]
            )
            self._some_protected_call(True, True, False)

    def test_incorrect_pin(self):
        self.setup_mnemonic_pin_passphrase()
        self.client.setup_debuglink(button=True, pin_correct=False)
        with pytest.raises(PinException):
            self._some_protected_call(False, True, False)

    def test_cancelled_pin(self):
        self.setup_mnemonic_pin_passphrase()
        self.client.setup_debuglink(button=True, pin_correct=False)  # PIN cancel
        with pytest.raises(PinException):
            self._some_protected_call(False, True, False)

    def test_exponential_backoff_with_reboot(self):
        self.setup_mnemonic_pin_passphrase()

        self.client.setup_debuglink(button=True, pin_correct=False)

        def test_backoff(attempts, start):
            if attempts <= 1:
                expected = 0
            else:
                expected = (2 ** (attempts - 1)) - 1
            got = round(time.time() - start, 2)

            msg = "Pin delay expected to be at least %s seconds, got %s" % (
                expected,
                got,
            )
            print(msg)
            assert got >= expected

        for attempt in range(1, 4):
            start = time.time()
            with pytest.raises(PinException):
                self._some_protected_call(False, True, False)
            test_backoff(attempt, start)



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

from trezorlib import btc, messages as proto
from trezorlib.tools import H_, parse_path

from ..support import ckd_public as bip32
from ..support.tx_cache import tx_cache
from .common import TrezorTest

TX_API = tx_cache("Testnet")


class TestMultisigChange(TrezorTest):
    node_ext1 = bip32.deserialize(
        "tpubDADHV9u9Y6gkggintTdMjJE3be58zKNLhpxBQyuEM6Pwx3sN9JVLmMCMN4DNVwL9AKec27z5TaWcWuHzMXiGAtcra5DjwWbvppGX4gaEGVN"
    )
    # m/1 => 02c0d0c5fee952620757c6128dbf327c996cd72ed3358d15d6518a1186099bc15e
    # m/2 => 0375b9dfaad928ce1a7eed88df7c084e67d99e9ab74332419458a9a45779706801

    node_ext2 = bip32.deserialize(
        "tpubDADHV9u9Y6gkhWXBmDJ6TUhZajLWjvKukRe2w9FfhdbQpUux8Z8jnPHNAZqFRgHPg9sR7YR93xThM32M7NfRu8S5WyDtext7S62sqxeJNkd"
    )
    # m/1 => 0388460dc439f4c8f5bcfc268c36e11b4375cad5c3535c336cfdf8c32c3afad5c1
    # m/2 => 03a04f945d5a3685729dde697d574076de4bdf38e904f813b22a851548e1110fc0

    node_ext3 = bip32.deserialize(
        "tpubDADHV9u9Y6gkmM5ohWRGTswrc6fr7soH7e2D2ic5a86PDUaHc5Ln9EbER69cEr5bDZPa7EXguJ1MhWVzPZpZWVdG5fvoF3hfirXvRbpCCBg"
    )
    # m/1 => 02e0c21e2a7cf00b94c5421725acff97f9826598b91f2340c5ddda730caca7d648
    # m/2 => 03928301ffb8c0d7a364b794914c716ba3107cc78a6fe581028b0d8638b22e8573

    node_int = bip32.deserialize(
        "tpubDADHV9u9Y6gke2Vw3rWE8KRXmeK8PTtsF5B3Cqjo6h3SoiyRtzxjnDVG1knxrqB8BpP1dMAd6MR3Ps5UXibiFDtQuWVPXLkJ3HvttZYbH12"
    )
    # m/1 => 03f91460d79e4e463d7d90cb75254bcd62b515a99a950574c721efdc5f711dff35
    # m/2 => 038caebd6f753bbbd2bb1f3346a43cd32140648583673a31d62f2dfb56ad0ab9e3

    # ext1 + ext2 + int
    #   redeemscript (2 of 3): 522102c0d0c5fee952620757c6128dbf327c996cd72ed3358d15d6518a1186099bc15e210388460dc439f4c8f5bcfc268c36e11b4375cad5c3535c336cfdf8c32c3afad5c1210338d78612e990f2eea0c426b5e48a8db70b9d7ed66282b3b26511e0b1c75515a653ae
    #   multisig address: 3Gj7y1FdTppx2JEDqYqAEZFnKCA4GRysKF
    #   tx: d1d08ea63255af4ad16b098e9885a252632086fa6be53301521d05253ce8a73d
    #   input 0: 0.001 BTC

    # ext1 + int + ext2
    #   redeemscript (2 of 3): 522102c0d0c5fee952620757c6128dbf327c996cd72ed3358d15d6518a1186099bc15e210338d78612e990f2eea0c426b5e48a8db70b9d7ed66282b3b26511e0b1c75515a6210388460dc439f4c8f5bcfc268c36e11b4375cad5c3535c336cfdf8c32c3afad5c153ae
    #   multisig address: 3QsvfB6d1LzYcpm8xyhS1N1HBRrzHTgLHB
    #   tx: a6e2829d089cee47e481b1a753a53081b40738cc87e38f1d9b23ab57d9ad4396
    #   input 0: 0.001 BTC

    # ext1 + ext3 + int
    #   redeemscript (2 of 3): 522102c0d0c5fee952620757c6128dbf327c996cd72ed3358d15d6518a1186099bc15e2102e0c21e2a7cf00b94c5421725acff97f9826598b91f2340c5ddda730caca7d648210338d78612e990f2eea0c426b5e48a8db70b9d7ed66282b3b26511e0b1c75515a653ae
    #   multisig address: 37LvC1Q5CyKbMbKMncEJdXxqGhHxrBEgPE
    #   tx: e4bc1ae5e5007a08f2b3926fe11c66612e8f73c6b00c69c7027213b84d259be3
    #   input 1: 0.001 BTC

    multisig_in1 = proto.MultisigRedeemScriptType(
        nodes=[node_ext2, node_ext1, node_int],
        address_n=[0, 0],
        signatures=[b"", b"", b""],
        m=2,
    )

    multisig_in2 = proto.MultisigRedeemScriptType(
        nodes=[node_ext1, node_ext2, node_int],
        address_n=[0, 1],
        signatures=[b"", b"", b""],
        m=2,
    )

    multisig_in3 = proto.MultisigRedeemScriptType(
        nodes=[node_ext1, node_ext3, node_int],
        address_n=[0, 1],
        signatures=[b"", b"", b""],
        m=2,
    )

    # 2N9W4z9AhAPaHghtqVQPbaTAGHdbrhKeBQw
    inp1 = proto.TxInputType(
        address_n=[H_(45), 0, 0, 0],
        prev_hash=bytes.fromhex(
            "16c6c8471b8db7a628f2b2bb86bfeefae1766463ce8692438c7fd3fce3f43ce5"
        ),
        prev_index=1,
        script_type=proto.InputScriptType.SPENDMULTISIG,
        multisig=multisig_in1,
    )

    # 2NDBG6QXQLtnQ3jRGkrqo53BiCeXfQXLdj4
    inp2 = proto.TxInputType(
        address_n=[H_(45), 0, 0, 1],
        prev_hash=bytes.fromhex(
            "d80c34ee14143a8bf61125102b7ef594118a3796cad670fa8ee15080ae155318"
        ),
        prev_index=0,
        script_type=proto.InputScriptType.SPENDMULTISIG,
        multisig=multisig_in2,
    )

    # 2MvwPWfp2XPU3S1cMwgEMKBPUw38VP5SBE4
    inp3 = proto.TxInputType(
        address_n=[H_(45), 0, 0, 1],
        prev_hash=bytes.fromhex(
            "b0946dc27ba308a749b11afecc2018980af18f79e89ad6b080b58220d856f739"
        ),
        prev_index=0,
        script_type=proto.InputScriptType.SPENDMULTISIG,
        multisig=multisig_in3,
    )

    def _responses(self, inp1, inp2, change=0):
        resp = [
            proto.TxRequest(
                request_type=proto.RequestType.TXINPUT,
                details=proto.TxRequestDetailsType(request_index=0),
            ),
            proto.TxRequest(
                request_type=proto.RequestType.TXMETA,
                details=proto.TxRequestDetailsType(tx_hash=inp1.prev_hash),
            ),
            proto.TxRequest(
                request_type=proto.RequestType.TXINPUT,
                details=proto.TxRequestDetailsType(
                    request_index=0, tx_hash=inp1.prev_hash
                ),
            ),
            proto.TxRequest(
                request_type=proto.RequestType.TXOUTPUT,
                details=proto.TxRequestDetailsType(
                    request_index=0, tx_hash=inp1.prev_hash
                ),
            ),
            proto.TxRequest(
                request_type=proto.RequestType.TXOUTPUT,
                details=proto.TxRequestDetailsType(
                    request_index=1, tx_hash=inp1.prev_hash
                ),
            ),
            proto.TxRequest(
                request_type=proto.RequestType.TXINPUT,
                details=proto.TxRequestDetailsType(request_index=1),
            ),
            proto.TxRequest(
                request_type=proto.RequestType.TXMETA,
                details=proto.TxRequestDetailsType(tx_hash=inp2.prev_hash),
            ),
            proto.TxRequest(
                request_type=proto.RequestType.TXINPUT,
                details=proto.TxRequestDetailsType(
                    request_index=0, tx_hash=inp2.prev_hash
                ),
            ),
            proto.TxRequest(
                request_type=proto.RequestType.TXOUTPUT,
                details=proto.TxRequestDetailsType(
                    request_index=0, tx_hash=inp2.prev_hash
                ),
            ),
            proto.TxRequest(
                request_type=proto.RequestType.TXOUTPUT,
                details=proto.TxRequestDetailsType(
                    request_index=1, tx_hash=inp2.prev_hash
                ),
            ),
            proto.TxRequest(
                request_type=proto.RequestType.TXOUTPUT,
                details=proto.TxRequestDetailsType(request_index=0),
            ),
        ]
        if change != 1:
            resp.append(proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput))
        resp.append(
            proto.TxRequest(
                request_type=proto.RequestType.TXOUTPUT,
                details=proto.TxRequestDetailsType(request_index=1),
            )
        )
        if change != 2:
            resp.append(proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput))
        resp += [
            proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
            proto.TxRequest(
                request_type=proto.RequestType.TXINPUT,
                details=proto.TxRequestDetailsType(request_index=0),
            ),
            proto.TxRequest(
                request_type=proto.RequestType.TXINPUT,
                details=proto.TxRequestDetailsType(request_index=1),
            ),
            proto.TxRequest(
                request_type=proto.RequestType.TXOUTPUT,
                details=proto.TxRequestDetailsType(request_index=0),
            ),
            proto.TxRequest(
                request_type=proto.RequestType.TXOUTPUT,
                details=proto.TxRequestDetailsType(request_index=1),
            ),
            proto.TxRequest(
                request_type=proto.RequestType.TXINPUT,
                details=proto.TxRequestDetailsType(request_index=0),
            ),
            proto.TxRequest(
                request_type=proto.RequestType.TXINPUT,
                details=proto.TxRequestDetailsType(request_index=1),
            ),
            proto.TxRequest(
                request_type=proto.RequestType.TXOUTPUT,
                details=proto.TxRequestDetailsType(request_index=0),
            ),
            proto.TxRequest(
                request_type=proto.RequestType.TXOUTPUT,
                details=proto.TxRequestDetailsType(request_index=1),
            ),
            proto.TxRequest(
                request_type=proto.RequestType.TXOUTPUT,
                details=proto.TxRequestDetailsType(request_index=0),
            ),
            proto.TxRequest(
                request_type=proto.RequestType.TXOUTPUT,
                details=proto.TxRequestDetailsType(request_index=1),
            ),
            proto.TxRequest(request_type=proto.RequestType.TXFINISHED),
        ]
        return resp

    # both outputs are external
    def test_external_external(self):
        self.setup_mnemonic_nopin_nopassphrase()

        out1 = proto.TxOutputType(
            address="muevUcG1Bb8eM2nGUGhqmeujHRX7YXjSEu",
            amount=40000000,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )

        out2 = proto.TxOutputType(
            address="mwdrpMVSJxxsM8f8xbnCHn9ERaRT1NG1UX",
            amount=44000000,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )

        with self.client:
            self.client.set_expected_responses(self._responses(self.inp1, self.inp2))
            _, serialized_tx = btc.sign_tx(
                self.client,
                "Testnet",
                [self.inp1, self.inp2],
                [out1, out2],
                prev_txes=TX_API,
            )

        assert (
            serialized_tx.hex()
            == "0100000002e53cf4e3fcd37f8c439286ce636476e1faeebf86bbb2f228a6b78d1b47c8c61601000000b400473044022064f13801744a6c21b694f62cdb5d834e852f13ecf85ed4d0a56ba279571c24e3022010fab4cb05bdd7b24c8376dda4f62a418548eea6eb483e58675fa06e0d5c642c014c69522103dc07026aacb5918dac4e09f9da8290d0ae22161699636c22cace78082116a7792103e70db185fad69c2971f0107a42930e5d82a9ed3a11b922a96fdfc4124b63e54c2103f3fe007a1e34ac76c1a2528e9149f90f9f93739929797afab6a8e18d682fa71053aeffffffff185315ae8050e18efa70d6ca96378a1194f57e2b102511f68b3a1414ee340cd800000000b4004730440220727b2522268f913acd213c507d7801b146e5b6cef666ad44b769c26d6c762e4d022021c0c2e9e8298dee2a490d956f7ab1b2d3160c1e37a50cc6d19a5e62eb484fc9014c6952210297ad8a5df42f9e362ef37d9a4ddced89d8f7a143690649aa0d0ff049c7daca842103ed1fd93989595d7ad4b488efd05a22c0239482c9a20923f2f214a38e54f6c41a2103f91460d79e4e463d7d90cb75254bcd62b515a99a950574c721efdc5f711dff3553aeffffffff02005a6202000000001976a9149b139230e4fe91c05a37ec334dc8378f3dbe377088ac00639f02000000001976a914b0d05a10926a7925508febdbab9a5bd4cda8c8f688ac00000000"
        )

    # first external, second internal
    def test_external_internal(self):
        self.setup_mnemonic_nopin_nopassphrase()

        out1 = proto.TxOutputType(
            address="muevUcG1Bb8eM2nGUGhqmeujHRX7YXjSEu",
            amount=40000000,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )

        out2 = proto.TxOutputType(
            address_n=parse_path("45'/0/1/1"),
            amount=44000000,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )

        with self.client:
            self.client.set_expected_responses(
                self._responses(self.inp1, self.inp2, change=2)
            )
            _, serialized_tx = btc.sign_tx(
                self.client,
                "Testnet",
                [self.inp1, self.inp2],
                [out1, out2],
                prev_txes=TX_API,
            )

        assert (
            serialized_tx.hex()
            == "0100000002e53cf4e3fcd37f8c439286ce636476e1faeebf86bbb2f228a6b78d1b47c8c61601000000b400473044022064f13801744a6c21b694f62cdb5d834e852f13ecf85ed4d0a56ba279571c24e3022010fab4cb05bdd7b24c8376dda4f62a418548eea6eb483e58675fa06e0d5c642c014c69522103dc07026aacb5918dac4e09f9da8290d0ae22161699636c22cace78082116a7792103e70db185fad69c2971f0107a42930e5d82a9ed3a11b922a96fdfc4124b63e54c2103f3fe007a1e34ac76c1a2528e9149f90f9f93739929797afab6a8e18d682fa71053aeffffffff185315ae8050e18efa70d6ca96378a1194f57e2b102511f68b3a1414ee340cd800000000b4004730440220727b2522268f913acd213c507d7801b146e5b6cef666ad44b769c26d6c762e4d022021c0c2e9e8298dee2a490d956f7ab1b2d3160c1e37a50cc6d19a5e62eb484fc9014c6952210297ad8a5df42f9e362ef37d9a4ddced89d8f7a143690649aa0d0ff049c7daca842103ed1fd93989595d7ad4b488efd05a22c0239482c9a20923f2f214a38e54f6c41a2103f91460d79e4e463d7d90cb75254bcd62b515a99a950574c721efdc5f711dff3553aeffffffff02005a6202000000001976a9149b139230e4fe91c05a37ec334dc8378f3dbe377088ac00639f02000000001976a914b0d05a10926a7925508febdbab9a5bd4cda8c8f688ac00000000"
        )

    # first internal, second external
    def test_internal_external(self):
        self.setup_mnemonic_nopin_nopassphrase()

        out1 = proto.TxOutputType(
            address_n=parse_path("45'/0/1/0"),
            amount=40000000,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )

        out2 = proto.TxOutputType(
            address="mwdrpMVSJxxsM8f8xbnCHn9ERaRT1NG1UX",
            amount=44000000,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )

        with self.client:
            self.client.set_expected_responses(
                self._responses(self.inp1, self.inp2, change=1)
            )
            _, serialized_tx = btc.sign_tx(
                self.client,
                "Testnet",
                [self.inp1, self.inp2],
                [out1, out2],
                prev_txes=TX_API,
            )

        assert (
            serialized_tx.hex()
            == "0100000002e53cf4e3fcd37f8c439286ce636476e1faeebf86bbb2f228a6b78d1b47c8c61601000000b400473044022064f13801744a6c21b694f62cdb5d834e852f13ecf85ed4d0a56ba279571c24e3022010fab4cb05bdd7b24c8376dda4f62a418548eea6eb483e58675fa06e0d5c642c014c69522103dc07026aacb5918dac4e09f9da8290d0ae22161699636c22cace78082116a7792103e70db185fad69c2971f0107a42930e5d82a9ed3a11b922a96fdfc4124b63e54c2103f3fe007a1e34ac76c1a2528e9149f90f9f93739929797afab6a8e18d682fa71053aeffffffff185315ae8050e18efa70d6ca96378a1194f57e2b102511f68b3a1414ee340cd800000000b4004730440220727b2522268f913acd213c507d7801b146e5b6cef666ad44b769c26d6c762e4d022021c0c2e9e8298dee2a490d956f7ab1b2d3160c1e37a50cc6d19a5e62eb484fc9014c6952210297ad8a5df42f9e362ef37d9a4ddced89d8f7a143690649aa0d0ff049c7daca842103ed1fd93989595d7ad4b488efd05a22c0239482c9a20923f2f214a38e54f6c41a2103f91460d79e4e463d7d90cb75254bcd62b515a99a950574c721efdc5f711dff3553aeffffffff02005a6202000000001976a9149b139230e4fe91c05a37ec334dc8378f3dbe377088ac00639f02000000001976a914b0d05a10926a7925508febdbab9a5bd4cda8c8f688ac00000000"
        )

    # both outputs are external
    def test_multisig_external_external(self):
        self.setup_mnemonic_nopin_nopassphrase()

        out1 = proto.TxOutputType(
            address="2N2aFoogGntQFFwdUVPfRmutXD22ThcNTsR",
            amount=40000000,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )

        out2 = proto.TxOutputType(
            address="2NFJjQcU8mw4Z3ywpbek8HL1VoJ27GDrkHw",
            amount=44000000,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )

        with self.client:
            self.client.set_expected_responses(self._responses(self.inp1, self.inp2))
            _, serialized_tx = btc.sign_tx(
                self.client,
                "Testnet",
                [self.inp1, self.inp2],
                [out1, out2],
                prev_txes=TX_API,
            )

        assert (
            serialized_tx.hex()
            == "0100000002e53cf4e3fcd37f8c439286ce636476e1faeebf86bbb2f228a6b78d1b47c8c61601000000b400473044022059394e0dfcb2d2f4a6108703f801545ca5a820c0ac6a1859d0a3854813de55fa02207b6a57d70b82932ff58163336c461653a2dc82c78ed8157159e5178ac7325390014c69522103dc07026aacb5918dac4e09f9da8290d0ae22161699636c22cace78082116a7792103e70db185fad69c2971f0107a42930e5d82a9ed3a11b922a96fdfc4124b63e54c2103f3fe007a1e34ac76c1a2528e9149f90f9f93739929797afab6a8e18d682fa71053aeffffffff185315ae8050e18efa70d6ca96378a1194f57e2b102511f68b3a1414ee340cd800000000b40047304402205a911685f5b974b2fc4a19d5ce056218773a4d20b5eaae2c2f9594929308182002201e03449f5a8813ec19f408bf1b6f4f334886d6fcf9920e300fd7678ef0724f81014c6952210297ad8a5df42f9e362ef37d9a4ddced89d8f7a143690649aa0d0ff049c7daca842103ed1fd93989595d7ad4b488efd05a22c0239482c9a20923f2f214a38e54f6c41a2103f91460d79e4e463d7d90cb75254bcd62b515a99a950574c721efdc5f711dff3553aeffffffff02005a62020000000017a91466528dd543f94d162c8111d2ec248d25ba9b90948700639f020000000017a914f1fc92c0aed1712911c70a2e09ac15ff0922652f8700000000"
        )

    # inputs match, change matches (first is change)
    def test_multisig_change_match_first(self):
        self.setup_mnemonic_nopin_nopassphrase()

        multisig_out1 = proto.MultisigRedeemScriptType(
            nodes=[self.node_ext2, self.node_ext1, self.node_int],
            address_n=[1, 0],
            signatures=[b"", b"", b""],
            m=2,
        )

        out1 = proto.TxOutputType(
            address_n=[H_(45), 0, 1, 0],
            multisig=multisig_out1,
            amount=40000000,
            script_type=proto.OutputScriptType.PAYTOMULTISIG,
        )

        out2 = proto.TxOutputType(
            address="2NFJjQcU8mw4Z3ywpbek8HL1VoJ27GDrkHw",
            amount=44000000,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )

        with self.client:
            self.client.set_expected_responses(
                self._responses(self.inp1, self.inp2, change=1)
            )
            _, serialized_tx = btc.sign_tx(
                self.client,
                "Testnet",
                [self.inp1, self.inp2],
                [out1, out2],
                prev_txes=TX_API,
            )

        assert (
            serialized_tx.hex()
            == "0100000002e53cf4e3fcd37f8c439286ce636476e1faeebf86bbb2f228a6b78d1b47c8c61601000000b400473044022059394e0dfcb2d2f4a6108703f801545ca5a820c0ac6a1859d0a3854813de55fa02207b6a57d70b82932ff58163336c461653a2dc82c78ed8157159e5178ac7325390014c69522103dc07026aacb5918dac4e09f9da8290d0ae22161699636c22cace78082116a7792103e70db185fad69c2971f0107a42930e5d82a9ed3a11b922a96fdfc4124b63e54c2103f3fe007a1e34ac76c1a2528e9149f90f9f93739929797afab6a8e18d682fa71053aeffffffff185315ae8050e18efa70d6ca96378a1194f57e2b102511f68b3a1414ee340cd800000000b40047304402205a911685f5b974b2fc4a19d5ce056218773a4d20b5eaae2c2f9594929308182002201e03449f5a8813ec19f408bf1b6f4f334886d6fcf9920e300fd7678ef0724f81014c6952210297ad8a5df42f9e362ef37d9a4ddced89d8f7a143690649aa0d0ff049c7daca842103ed1fd93989595d7ad4b488efd05a22c0239482c9a20923f2f214a38e54f6c41a2103f91460d79e4e463d7d90cb75254bcd62b515a99a950574c721efdc5f711dff3553aeffffffff02005a62020000000017a91466528dd543f94d162c8111d2ec248d25ba9b90948700639f020000000017a914f1fc92c0aed1712911c70a2e09ac15ff0922652f8700000000"
        )

    # inputs match, change matches (second is change)
    def test_multisig_change_match_second(self):
        self.setup_mnemonic_nopin_nopassphrase()

        multisig_out2 = proto.MultisigRedeemScriptType(
            nodes=[self.node_ext1, self.node_ext2, self.node_int],
            address_n=[1, 1],
            signatures=[b"", b"", b""],
            m=2,
        )

        out1 = proto.TxOutputType(
            address="2N2aFoogGntQFFwdUVPfRmutXD22ThcNTsR",
            amount=40000000,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )

        out2 = proto.TxOutputType(
            address_n=[H_(45), 0, 1, 1],
            multisig=multisig_out2,
            amount=44000000,
            script_type=proto.OutputScriptType.PAYTOMULTISIG,
        )

        with self.client:
            self.client.set_expected_responses(
                self._responses(self.inp1, self.inp2, change=2)
            )
            _, serialized_tx = btc.sign_tx(
                self.client,
                "Testnet",
                [self.inp1, self.inp2],
                [out1, out2],
                prev_txes=TX_API,
            )

        assert (
            serialized_tx.hex()
            == "0100000002e53cf4e3fcd37f8c439286ce636476e1faeebf86bbb2f228a6b78d1b47c8c61601000000b400473044022059394e0dfcb2d2f4a6108703f801545ca5a820c0ac6a1859d0a3854813de55fa02207b6a57d70b82932ff58163336c461653a2dc82c78ed8157159e5178ac7325390014c69522103dc07026aacb5918dac4e09f9da8290d0ae22161699636c22cace78082116a7792103e70db185fad69c2971f0107a42930e5d82a9ed3a11b922a96fdfc4124b63e54c2103f3fe007a1e34ac76c1a2528e9149f90f9f93739929797afab6a8e18d682fa71053aeffffffff185315ae8050e18efa70d6ca96378a1194f57e2b102511f68b3a1414ee340cd800000000b40047304402205a911685f5b974b2fc4a19d5ce056218773a4d20b5eaae2c2f9594929308182002201e03449f5a8813ec19f408bf1b6f4f334886d6fcf9920e300fd7678ef0724f81014c6952210297ad8a5df42f9e362ef37d9a4ddced89d8f7a143690649aa0d0ff049c7daca842103ed1fd93989595d7ad4b488efd05a22c0239482c9a20923f2f214a38e54f6c41a2103f91460d79e4e463d7d90cb75254bcd62b515a99a950574c721efdc5f711dff3553aeffffffff02005a62020000000017a91466528dd543f94d162c8111d2ec248d25ba9b90948700639f020000000017a914f1fc92c0aed1712911c70a2e09ac15ff0922652f8700000000"
        )

    # inputs match, change mismatches (second tries to be change but isn't)
    def test_multisig_mismatch_change(self):
        self.setup_mnemonic_nopin_nopassphrase()

        multisig_out2 = proto.MultisigRedeemScriptType(
            nodes=[self.node_ext1, self.node_int, self.node_ext3],
            address_n=[1, 0],
            signatures=[b"", b"", b""],
            m=2,
        )

        out1 = proto.TxOutputType(
            address="2N2aFoogGntQFFwdUVPfRmutXD22ThcNTsR",
            amount=40000000,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )

        out2 = proto.TxOutputType(
            address_n=[H_(45), 0, 1, 0],
            multisig=multisig_out2,
            amount=44000000,
            script_type=proto.OutputScriptType.PAYTOMULTISIG,
        )

        with self.client:
            self.client.set_expected_responses(self._responses(self.inp1, self.inp2))
            _, serialized_tx = btc.sign_tx(
                self.client,
                "Testnet",
                [self.inp1, self.inp2],
                [out1, out2],
                prev_txes=TX_API,
            )

        assert (
            serialized_tx.hex()
            == "0100000002e53cf4e3fcd37f8c439286ce636476e1faeebf86bbb2f228a6b78d1b47c8c61601000000b40047304402207f9992cc0230527faf54ec6bd233307db82bc8fac039dcee418bc6feb4e96a3a02206bb4cb157ad27c123277328a877572563a45d70b844d9ab07cc42238112f8c2a014c69522103dc07026aacb5918dac4e09f9da8290d0ae22161699636c22cace78082116a7792103e70db185fad69c2971f0107a42930e5d82a9ed3a11b922a96fdfc4124b63e54c2103f3fe007a1e34ac76c1a2528e9149f90f9f93739929797afab6a8e18d682fa71053aeffffffff185315ae8050e18efa70d6ca96378a1194f57e2b102511f68b3a1414ee340cd800000000b400473044022078a41bfa87d72d6ba810d84bf568b5a29acf8b851ba6c3a8dbff079b34a7feb0022037b770c776db0b6c883c38a684a121b90a59ed1958774cbf64de70e53e29639f014c6952210297ad8a5df42f9e362ef37d9a4ddced89d8f7a143690649aa0d0ff049c7daca842103ed1fd93989595d7ad4b488efd05a22c0239482c9a20923f2f214a38e54f6c41a2103f91460d79e4e463d7d90cb75254bcd62b515a99a950574c721efdc5f711dff3553aeffffffff02005a62020000000017a91466528dd543f94d162c8111d2ec248d25ba9b90948700639f020000000017a914e6a3e2fbadb7f559f8d20c46aceae78c96fcf1d18700000000"
        )

    # inputs mismatch, change matches with first input
    def test_multisig_mismatch_inputs(self):
        self.setup_mnemonic_nopin_nopassphrase()

        multisig_out1 = proto.MultisigRedeemScriptType(
            nodes=[self.node_ext2, self.node_ext1, self.node_int],
            address_n=[1, 0],
            signatures=[b"", b"", b""],
            m=2,
        )

        out1 = proto.TxOutputType(
            address_n=[H_(45), 0, 1, 0],
            multisig=multisig_out1,
            amount=40000000,
            script_type=proto.OutputScriptType.PAYTOMULTISIG,
        )

        out2 = proto.TxOutputType(
            address="2NFJjQcU8mw4Z3ywpbek8HL1VoJ27GDrkHw",
            amount=65000000,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )

        with self.client:
            self.client.set_expected_responses(self._responses(self.inp1, self.inp3))
            _, serialized_tx = btc.sign_tx(
                self.client,
                "Testnet",
                [self.inp1, self.inp3],
                [out1, out2],
                prev_txes=TX_API,
            )

        assert (
            serialized_tx.hex()
            == "0100000002e53cf4e3fcd37f8c439286ce636476e1faeebf86bbb2f228a6b78d1b47c8c61601000000b500483045022100d907b9339951c96ef4515ef7aff8b3c28c4c8c5875d7421aa1de9f3a94e3508302205cdc311a6c91dfbb74f1a9a940a994a65dbfb0cf6dedcaaaeee839e0b8fd016d014c69522103dc07026aacb5918dac4e09f9da8290d0ae22161699636c22cace78082116a7792103e70db185fad69c2971f0107a42930e5d82a9ed3a11b922a96fdfc4124b63e54c2103f3fe007a1e34ac76c1a2528e9149f90f9f93739929797afab6a8e18d682fa71053aeffffffff39f756d82082b580b0d69ae8798ff10a981820ccfe1ab149a708a37bc26d94b000000000b500483045022100fdad4a47d15f47cc364fe0cbed11b1ced1f9ef210bc1bd413ec4384f630c63720220752e4f09ea4e5e6623f5ebe89b3983ec6e5702f63f9bce696f10b2d594d23532014c6952210297ad8a5df42f9e362ef37d9a4ddced89d8f7a143690649aa0d0ff049c7daca842103b6321a1194e5cc47b6b7edc3f67a096e6f71ccb72440f84f390b6e98df0ea8ec2103f91460d79e4e463d7d90cb75254bcd62b515a99a950574c721efdc5f711dff3553aeffffffff02005a62020000000017a91466528dd543f94d162c8111d2ec248d25ba9b90948740d2df030000000017a914f1fc92c0aed1712911c70a2e09ac15ff0922652f8700000000"
        )



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None  # type: ignore


class LiskMultisignatureType(p.MessageType):

    def __init__(
        self,
        min: int = None,
        life_time: int = None,
        keys_group: List[str] = None,
    ) -> None:
        self.min = min
        self.life_time = life_time
        self.keys_group = keys_group if keys_group is not None else []

    @classmethod
    def get_fields(cls):
        return {
            1: ('min', p.UVarintType, 0),
            2: ('life_time', p.UVarintType, 0),
            3: ('keys_group', p.UnicodeType, p.FLAG_REPEATED),
        }



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import json

from . import messages as proto
from .tools import CallException, expect

TYPE_TRANSACTION_TRANSFER = 0x0101
TYPE_IMPORTANCE_TRANSFER = 0x0801
TYPE_AGGREGATE_MODIFICATION = 0x1001
TYPE_MULTISIG_SIGNATURE = 0x1002
TYPE_MULTISIG = 0x1004
TYPE_PROVISION_NAMESPACE = 0x2001
TYPE_MOSAIC_CREATION = 0x4001
TYPE_MOSAIC_SUPPLY_CHANGE = 0x4002


def create_transaction_common(transaction):
    msg = proto.NEMTransactionCommon()
    msg.network = (transaction["version"] >> 24) & 0xFF
    msg.timestamp = transaction["timeStamp"]
    msg.fee = transaction["fee"]
    msg.deadline = transaction["deadline"]

    if "signer" in transaction:
        msg.signer = bytes.fromhex(transaction["signer"])

    return msg


def create_transfer(transaction):
    msg = proto.NEMTransfer()
    msg.recipient = transaction["recipient"]
    msg.amount = transaction["amount"]

    if "payload" in transaction["message"]:
        msg.payload = bytes.fromhex(transaction["message"]["payload"])

        if transaction["message"]["type"] == 0x02:
            msg.public_key = bytes.fromhex(transaction["message"]["publicKey"])

    if "mosaics" in transaction:
        msg.mosaics = [
            proto.NEMMosaic(
                namespace=mosaic["mosaicId"]["namespaceId"],
                mosaic=mosaic["mosaicId"]["name"],
                quantity=mosaic["quantity"],
            )
            for mosaic in transaction["mosaics"]
        ]

    return msg


def create_aggregate_modification(transactions):
    msg = proto.NEMAggregateModification()
    msg.modifications = [
        proto.NEMCosignatoryModification(
            type=modification["modificationType"],
            public_key=bytes.fromhex(modification["cosignatoryAccount"]),
        )
        for modification in transactions["modifications"]
    ]

    if "minCosignatories" in transactions:
        msg.relative_change = transactions["minCosignatories"]["relativeChange"]

    return msg


def create_provision_namespace(transaction):
    msg = proto.NEMProvisionNamespace()
    msg.namespace = transaction["newPart"]

    if transaction["parent"]:
        msg.parent = transaction["parent"]

    msg.sink = transaction["rentalFeeSink"]
    msg.fee = transaction["rentalFee"]
    return msg


def create_mosaic_creation(transaction):
    definition = transaction["mosaicDefinition"]
    msg = proto.NEMMosaicCreation()
    msg.definition = proto.NEMMosaicDefinition()
    msg.definition.namespace = definition["id"]["namespaceId"]
    msg.definition.mosaic = definition["id"]["name"]

    if definition["levy"]:
        msg.definition.levy = definition["levy"]["type"]
        msg.definition.fee = definition["levy"]["fee"]
        msg.definition.levy_address = definition["levy"]["recipient"]
        msg.definition.levy_namespace = definition["levy"]["mosaicId"]["namespaceId"]
        msg.definition.levy_mosaic = definition["levy"]["mosaicId"]["name"]

    msg.definition.description = definition["description"]

    for property in definition["properties"]:
        name = property["name"]
        value = json.loads(property["value"])

        if name == "divisibility":
            msg.definition.divisibility = value
        elif name == "initialSupply":
            msg.definition.supply = value
        elif name == "supplyMutable":
            msg.definition.mutable_supply = value
        elif name == "transferable":
            msg.definition.transferable = value

    msg.sink = transaction["creationFeeSink"]
    msg.fee = transaction["creationFee"]
    return msg


def create_supply_change(transaction):
    msg = proto.NEMMosaicSupplyChange()
    msg.namespace = transaction["mosaicId"]["namespaceId"]
    msg.mosaic = transaction["mosaicId"]["name"]
    msg.type = transaction["supplyType"]
    msg.delta = transaction["delta"]
    return msg


def create_importance_transfer(transaction):
    msg = proto.NEMImportanceTransfer()
    msg.mode = transaction["importanceTransfer"]["mode"]
    msg.public_key = bytes.fromhex(transaction["importanceTransfer"]["publicKey"])
    return msg


def fill_transaction_by_type(msg, transaction):
    if transaction["type"] == TYPE_TRANSACTION_TRANSFER:
        msg.transfer = create_transfer(transaction)
    elif transaction["type"] == TYPE_AGGREGATE_MODIFICATION:
        msg.aggregate_modification = create_aggregate_modification(transaction)
    elif transaction["type"] == TYPE_PROVISION_NAMESPACE:
        msg.provision_namespace = create_provision_namespace(transaction)
    elif transaction["type"] == TYPE_MOSAIC_CREATION:
        msg.mosaic_creation = create_mosaic_creation(transaction)
    elif transaction["type"] == TYPE_MOSAIC_SUPPLY_CHANGE:
        msg.supply_change = create_supply_change(transaction)
    elif transaction["type"] == TYPE_IMPORTANCE_TRANSFER:
        msg.importance_transfer = create_importance_transfer(transaction)
    else:
        raise ValueError("Unknown transaction type")


def create_sign_tx(transaction):
    msg = proto.NEMSignTx()
    msg.transaction = create_transaction_common(transaction)
    msg.cosigning = transaction["type"] == TYPE_MULTISIG_SIGNATURE

    if transaction["type"] in (TYPE_MULTISIG_SIGNATURE, TYPE_MULTISIG):
        other_trans = transaction["otherTrans"]
        msg.multisig = create_transaction_common(other_trans)
        fill_transaction_by_type(msg, other_trans)
    elif "otherTrans" in transaction:
        raise ValueError("Transaction does not support inner transaction")
    else:
        fill_transaction_by_type(msg, transaction)

    return msg


# ====== Client functions ====== #


@expect(proto.NEMAddress, field="address")
def get_address(client, n, network, show_display=False):
    return client.call(
        proto.NEMGetAddress(address_n=n, network=network, show_display=show_display)
    )


@expect(proto.NEMSignedTx)
def sign_tx(client, n, transaction):
    try:
        msg = create_sign_tx(transaction)
    except ValueError as e:
        raise CallException(e.args)

    assert msg.transaction is not None
    msg.transaction.address_n = n
    return client.call(msg)



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class ResetDevice(p.MessageType):
    MESSAGE_WIRE_TYPE = 14

    def __init__(
        self,
        display_random: bool = None,
        strength: int = None,
        passphrase_protection: bool = None,
        pin_protection: bool = None,
        language: str = None,
        label: str = None,
        u2f_counter: int = None,
        skip_backup: bool = None,
        no_backup: bool = None,
    ) -> None:
        self.display_random = display_random
        self.strength = strength
        self.passphrase_protection = passphrase_protection
        self.pin_protection = pin_protection
        self.language = language
        self.label = label
        self.u2f_counter = u2f_counter
        self.skip_backup = skip_backup
        self.no_backup = no_backup

    @classmethod
    def get_fields(cls):
        return {
            1: ('display_random', p.BoolType, 0),
            2: ('strength', p.UVarintType, 0),  # default=256
            3: ('passphrase_protection', p.BoolType, 0),
            4: ('pin_protection', p.BoolType, 0),
            5: ('language', p.UnicodeType, 0),  # default=english
            6: ('label', p.UnicodeType, 0),
            7: ('u2f_counter', p.UVarintType, 0),
            8: ('skip_backup', p.BoolType, 0),
            9: ('no_backup', p.BoolType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class ECDHSessionKey(p.MessageType):
    MESSAGE_WIRE_TYPE = 62

    def __init__(
        self,
        session_key: bytes = None,
    ) -> None:
        self.session_key = session_key

    @classmethod
    def get_fields(cls):
        return {
            1: ('session_key', p.BytesType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .HDNodePathType import HDNodePathType
from .HDNodeType import HDNodeType

if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None  # type: ignore


class MultisigRedeemScriptType(p.MessageType):

    def __init__(
        self,
        pubkeys: List[HDNodePathType] = None,
        signatures: List[bytes] = None,
        m: int = None,
        nodes: List[HDNodeType] = None,
        address_n: List[int] = None,
    ) -> None:
        self.pubkeys = pubkeys if pubkeys is not None else []
        self.signatures = signatures if signatures is not None else []
        self.m = m
        self.nodes = nodes if nodes is not None else []
        self.address_n = address_n if address_n is not None else []

    @classmethod
    def get_fields(cls):
        return {
            1: ('pubkeys', HDNodePathType, p.FLAG_REPEATED),
            2: ('signatures', p.BytesType, p.FLAG_REPEATED),
            3: ('m', p.UVarintType, 0),
            4: ('nodes', HDNodeType, p.FLAG_REPEATED),
            5: ('address_n', p.UVarintType, p.FLAG_REPEATED),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class SignedIdentity(p.MessageType):
    MESSAGE_WIRE_TYPE = 54

    def __init__(
        self,
        address: str = None,
        public_key: bytes = None,
        signature: bytes = None,
    ) -> None:
        self.address = address
        self.public_key = public_key
        self.signature = signature

    @classmethod
    def get_fields(cls):
        return {
            1: ('address', p.UnicodeType, 0),
            2: ('public_key', p.BytesType, 0),
            3: ('signature', p.BytesType, 0),
        }



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

from copy import deepcopy

from mnemonic import Mnemonic

from . import messages as proto, protobuf, tools
from .client import TrezorClient
from .tools import expect

EXPECTED_RESPONSES_CONTEXT_LINES = 3


class DebugLink:
    def __init__(self, transport, auto_interact=True):
        self.transport = transport
        self.allow_interactions = auto_interact

    def open(self):
        self.transport.begin_session()

    def close(self):
        self.transport.end_session()

    def _call(self, msg, nowait=False):
        self.transport.write(msg)
        if nowait:
            return None
        ret = self.transport.read()
        return ret

    def state(self):
        return self._call(proto.DebugLinkGetState())

    def read_pin(self):
        state = self.state()
        return state.pin, state.matrix

    def read_pin_encoded(self):
        return self.encode_pin(*self.read_pin())

    def encode_pin(self, pin, matrix=None):
        """Transform correct PIN according to the displayed matrix."""
        if matrix is None:
            _, matrix = self.read_pin()
        return "".join([str(matrix.index(p) + 1) for p in pin])

    def read_layout(self):
        obj = self._call(proto.DebugLinkGetState())
        return obj.layout

    def read_mnemonic_secret(self):
        obj = self._call(proto.DebugLinkGetState())
        return obj.mnemonic_secret

    def read_recovery_word(self):
        obj = self._call(proto.DebugLinkGetState())
        return (obj.recovery_fake_word, obj.recovery_word_pos)

    def read_reset_word(self):
        obj = self._call(proto.DebugLinkGetState())
        return obj.reset_word

    def read_reset_word_pos(self):
        obj = self._call(proto.DebugLinkGetState())
        return obj.reset_word_pos

    def read_reset_entropy(self):
        obj = self._call(proto.DebugLinkGetState())
        return obj.reset_entropy

    def read_passphrase_protection(self):
        obj = self._call(proto.DebugLinkGetState())
        return obj.passphrase_protection

    def input(self, word=None, button=None, swipe=None):
        if not self.allow_interactions:
            return
        decision = proto.DebugLinkDecision()
        if button is not None:
            decision.yes_no = button
        elif word is not None:
            decision.input = word
        elif swipe is not None:
            decision.up_down = swipe
        else:
            raise ValueError("You need to provide input data.")
        self._call(decision, nowait=True)

    def press_button(self, yes_no):
        self._call(proto.DebugLinkDecision(yes_no=yes_no), nowait=True)

    def press_yes(self):
        self.input(button=True)

    def press_no(self):
        self.input(button=False)

    def swipe_up(self):
        self.input(swipe=True)

    def swipe_down(self):
        self.input(swipe=False)

    def stop(self):
        self._call(proto.DebugLinkStop(), nowait=True)

    @expect(proto.DebugLinkMemory, field="memory")
    def memory_read(self, address, length):
        return self._call(proto.DebugLinkMemoryRead(address=address, length=length))

    def memory_write(self, address, memory, flash=False):
        self._call(
            proto.DebugLinkMemoryWrite(address=address, memory=memory, flash=flash),
            nowait=True,
        )

    def flash_erase(self, sector):
        self._call(proto.DebugLinkFlashErase(sector=sector), nowait=True)


class NullDebugLink(DebugLink):
    def __init__(self):
        super().__init__(None)

    def open(self):
        pass

    def close(self):
        pass

    def _call(self, msg, nowait=False):
        if not nowait:
            if isinstance(msg, proto.DebugLinkGetState):
                return proto.DebugLinkState()
            else:
                raise RuntimeError("unexpected call to a fake debuglink")


class DebugUI:
    INPUT_FLOW_DONE = object()

    def __init__(self, debuglink: DebugLink):
        self.debuglink = debuglink
        self.pin = None
        self.passphrase = "sphinx of black quartz, judge my wov"
        self.input_flow = None

    def button_request(self, code):
        if self.input_flow is None:
            self.debuglink.press_yes()
        elif self.input_flow is self.INPUT_FLOW_DONE:
            raise AssertionError("input flow ended prematurely")
        else:
            try:
                self.input_flow.send(code)
            except StopIteration:
                self.input_flow = self.INPUT_FLOW_DONE

    def get_pin(self, code=None):
        if self.pin:
            return self.pin
        else:
            return self.debuglink.read_pin_encoded()

    def get_passphrase(self):
        return self.passphrase


class TrezorClientDebugLink(TrezorClient):
    # This class implements automatic responses
    # and other functionality for unit tests
    # for various callbacks, created in order
    # to automatically pass unit tests.
    #
    # This mixing should be used only for purposes
    # of unit testing, because it will fail to work
    # without special DebugLink interface provided
    # by the device.

    def __init__(self, transport, auto_interact=True):
        try:
            debug_transport = transport.find_debug()
            self.debug = DebugLink(debug_transport, auto_interact)
        except Exception:
            if not auto_interact:
                self.debug = NullDebugLink()
            else:
                raise

        self.ui = DebugUI(self.debug)

        self.in_with_statement = 0
        self.screenshot_id = 0

        self.filters = {}

        # Always press Yes and provide correct pin
        self.setup_debuglink(True, True)

        # Do not expect any specific response from device
        self.expected_responses = None
        self.current_response = None

        # Use blank passphrase
        self.set_passphrase("")
        super().__init__(transport, ui=self.ui)

    def open(self):
        super().open()
        self.debug.open()

    def close(self):
        self.debug.close()
        super().close()

    def set_filter(self, message_type, callback):
        self.filters[message_type] = callback

    def _filter_message(self, msg):
        message_type = msg.__class__
        callback = self.filters.get(message_type)
        if callable(callback):
            return callback(deepcopy(msg))
        else:
            return msg

    def set_input_flow(self, input_flow):
        if input_flow is None:
            self.ui.input_flow = None
            return

        if callable(input_flow):
            input_flow = input_flow()
        if not hasattr(input_flow, "send"):
            raise RuntimeError("input_flow should be a generator function")
        self.ui.input_flow = input_flow
        next(input_flow)  # can't send before first yield

    def __enter__(self):
        # For usage in with/expected_responses
        self.in_with_statement += 1
        return self

    def __exit__(self, _type, value, traceback):
        self.in_with_statement -= 1

        if _type is not None:
            # Another exception raised
            return False

        if self.expected_responses is None:
            # no need to check anything else
            return False

        # return isinstance(value, TypeError)
        # Evaluate missed responses in 'with' statement
        if self.current_response < len(self.expected_responses):
            self._raise_unexpected_response(None)

        # Cleanup
        self.expected_responses = None
        self.current_response = None
        return False

    def set_expected_responses(self, expected):
        if not self.in_with_statement:
            raise RuntimeError("Must be called inside 'with' statement")
        self.expected_responses = expected
        self.current_response = 0

    def setup_debuglink(self, button, pin_correct):
        # self.button = button  # True -> YES button, False -> NO button
        if pin_correct:
            self.ui.pin = None
        else:
            self.ui.pin = "444222"

    def set_passphrase(self, passphrase):
        self.ui.passphrase = Mnemonic.normalize_string(passphrase)

    def set_mnemonic(self, mnemonic):
        self.mnemonic = Mnemonic.normalize_string(mnemonic).split(" ")

    def _raw_read(self):
        __tracebackhide__ = True  # for pytest # pylint: disable=W0612

        # if SCREENSHOT and self.debug:
        #     from PIL import Image

        #     layout = self.debug.state().layout
        #     im = Image.new("RGB", (128, 64))
        #     pix = im.load()
        #     for x in range(128):
        #         for y in range(64):
        #             rx, ry = 127 - x, 63 - y
        #             if (ord(layout[rx + (ry / 8) * 128]) & (1 << (ry % 8))) > 0:
        #                 pix[x, y] = (255, 255, 255)
        #     im.save("scr%05d.png" % self.screenshot_id)
        #     self.screenshot_id += 1

        resp = super()._raw_read()
        resp = self._filter_message(resp)
        self._check_request(resp)
        return resp

    def _raw_write(self, msg):
        return super()._raw_write(self._filter_message(msg))

    def _raise_unexpected_response(self, msg):
        __tracebackhide__ = True  # for pytest # pylint: disable=W0612

        start_at = max(self.current_response - EXPECTED_RESPONSES_CONTEXT_LINES, 0)
        stop_at = min(
            self.current_response + EXPECTED_RESPONSES_CONTEXT_LINES + 1,
            len(self.expected_responses),
        )
        output = []
        output.append("Expected responses:")
        if start_at > 0:
            output.append("    (...{} previous responses omitted)".format(start_at))
        for i in range(start_at, stop_at):
            exp = self.expected_responses[i]
            prefix = "    " if i != self.current_response else ">>> "
            set_fields = {
                key: value
                for key, value in exp.__dict__.items()
                if value is not None and value != []
            }
            oneline_str = ", ".join("{}={!r}".format(*i) for i in set_fields.items())
            if len(oneline_str) < 60:
                output.append(
                    "{}{}({})".format(prefix, exp.__class__.__name__, oneline_str)
                )
            else:
                item = []
                item.append("{}{}(".format(prefix, exp.__class__.__name__))
                for key, value in set_fields.items():
                    item.append("{}    {}={!r}".format(prefix, key, value))
                item.append("{})".format(prefix))
                output.append("\n".join(item))
        if stop_at < len(self.expected_responses):
            omitted = len(self.expected_responses) - stop_at
            output.append("    (...{} following responses omitted)".format(omitted))

        output.append("")
        if msg is not None:
            output.append("Actually received:")
            output.append(protobuf.format_message(msg))
        else:
            output.append("This message was never received.")
        raise AssertionError("\n".join(output))

    def _check_request(self, msg):
        __tracebackhide__ = True  # for pytest # pylint: disable=W0612
        if self.expected_responses is None:
            return

        if self.current_response >= len(self.expected_responses):
            raise AssertionError(
                "No more messages were expected, but we got:\n"
                + protobuf.format_message(msg)
            )

        expected = self.expected_responses[self.current_response]

        if msg.__class__ != expected.__class__:
            self._raise_unexpected_response(msg)

        for field, value in expected.__dict__.items():
            if value is None or value == []:
                continue
            if getattr(msg, field) != value:
                self._raise_unexpected_response(msg)

        self.current_response += 1

    def mnemonic_callback(self, _):
        word, pos = self.debug.read_recovery_word()
        if word != "":
            return word
        if pos != 0:
            return self.mnemonic[pos - 1]

        raise RuntimeError("Unexpected call")


@expect(proto.Success, field="message")
def load_device_by_mnemonic(
    client,
    mnemonic,
    pin,
    passphrase_protection,
    label,
    language="english",
    skip_checksum=False,
    expand=False,
):
    # Convert mnemonic to UTF8 NKFD
    mnemonic = Mnemonic.normalize_string(mnemonic)

    # Convert mnemonic to ASCII stream
    mnemonic = mnemonic.encode()

    m = Mnemonic("english")

    if expand:
        mnemonic = m.expand(mnemonic)

    if not skip_checksum and not m.check(mnemonic):
        raise ValueError("Invalid mnemonic checksum")

    if client.features.initialized:
        raise RuntimeError(
            "Device is initialized already. Call device.wipe() and try again."
        )

    resp = client.call(
        proto.LoadDevice(
            mnemonic=mnemonic,
            pin=pin,
            passphrase_protection=passphrase_protection,
            language=language,
            label=label,
            skip_checksum=skip_checksum,
        )
    )
    client.init_device()
    return resp


@expect(proto.Success, field="message")
def load_device_by_xprv(client, xprv, pin, passphrase_protection, label, language):
    if client.features.initialized:
        raise RuntimeError(
            "Device is initialized already. Call wipe_device() and try again."
        )

    if xprv[0:4] not in ("xprv", "tprv"):
        raise ValueError("Unknown type of xprv")

    if not 100 < len(xprv) < 112:  # yes this is correct in Python
        raise ValueError("Invalid length of xprv")

    node = proto.HDNodeType()
    data = tools.b58decode(xprv, None).hex()

    if data[90:92] != "00":
        raise ValueError("Contain invalid private key")

    checksum = (tools.btc_hash(bytes.fromhex(data[:156]))[:4]).hex()
    if checksum != data[156:]:
        raise ValueError("Checksum doesn't match")

    # version 0488ade4
    # depth 00
    # fingerprint 00000000
    # child_num 00000000
    # chaincode 873dff81c02f525623fd1fe5167eac3a55a049de3d314bb42ee227ffed37d508
    # privkey   00e8f32e723decf4051aefac8e2c93c9c5b214313817cdb01a1494b917c8436b35
    # checksum e77e9d71

    node.depth = int(data[8:10], 16)
    node.fingerprint = int(data[10:18], 16)
    node.child_num = int(data[18:26], 16)
    node.chain_code = bytes.fromhex(data[26:90])
    node.private_key = bytes.fromhex(data[92:156])  # skip 0x00 indicating privkey

    resp = client.call(
        proto.LoadDevice(
            node=node,
            pin=pin,
            passphrase_protection=passphrase_protection,
            language=language,
            label=label,
        )
    )
    client.init_device()
    return resp


@expect(proto.Success, field="message")
def self_test(client):
    if client.features.bootloader_mode is not True:
        raise RuntimeError("Device must be in bootloader mode")

    return client.call(
        proto.SelfTest(
            payload=b"\x00\xFF\x55\xAA\x66\x99\x33\xCCABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!\x00\xFF\x55\xAA\x66\x99\x33\xCC"
        )
    )



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import pytest

from trezorlib import messages, ripple
from trezorlib.tools import CallException, parse_path

from .common import TrezorTest


@pytest.mark.ripple
@pytest.mark.skip_t1  # T1 support is not planned
class TestMsgRippleSignTx(TrezorTest):
    def test_ripple_sign_simple_tx(self):
        self.setup_mnemonic_allallall()

        msg = ripple.create_sign_tx_msg(
            {
                "TransactionType": "Payment",
                "Payment": {
                    "Amount": 100000000,
                    "Destination": "rBKz5MC2iXdoS3XgnNSYmF69K1Yo4NS3Ws",
                },
                "Flags": 0x80000000,
                "Fee": 100000,
                "Sequence": 25,
            }
        )
        resp = ripple.sign_tx(self.client, parse_path("m/44'/144'/0'/0/0"), msg)
        assert (
            resp.signature.hex()
            == "3045022100e243ef623675eeeb95965c35c3e06d63a9fc68bb37e17dc87af9c0af83ec057e02206ca8aa5eaab8396397aef6d38d25710441faf7c79d292ee1d627df15ad9346c0"
        )
        assert (
            resp.serialized_tx.hex()
            == "12000022800000002400000019614000000005f5e1006840000000000186a0732102131facd1eab748d6cddc492f54b04e8c35658894f4add2232ebc5afe7521dbe474473045022100e243ef623675eeeb95965c35c3e06d63a9fc68bb37e17dc87af9c0af83ec057e02206ca8aa5eaab8396397aef6d38d25710441faf7c79d292ee1d627df15ad9346c081148fb40e1ffa5d557ce9851a535af94965e0dd098883147148ebebf7304ccdf1676fefcf9734cf1e780826"
        )

        msg = ripple.create_sign_tx_msg(
            {
                "TransactionType": "Payment",
                "Payment": {
                    "Amount": 1,
                    "Destination": "rNaqKtKrMSwpwZSzRckPf7S96DkimjkF4H",
                },
                "Fee": 10,
                "Sequence": 1,
            }
        )
        resp = ripple.sign_tx(self.client, parse_path("m/44'/144'/0'/0/2"), msg)
        assert (
            resp.signature.hex()
            == "3044022069900e6e578997fad5189981b74b16badc7ba8b9f1052694033fa2779113ddc002206c8006ada310edf099fb22c0c12073550c8fc73247b236a974c5f1144831dd5f"
        )
        assert (
            resp.serialized_tx.hex()
            == "1200002280000000240000000161400000000000000168400000000000000a732103dbed1e77cb91a005e2ec71afbccce5444c9be58276665a3859040f692de8fed274463044022069900e6e578997fad5189981b74b16badc7ba8b9f1052694033fa2779113ddc002206c8006ada310edf099fb22c0c12073550c8fc73247b236a974c5f1144831dd5f8114bdf86f3ae715ba346b7772ea0e133f48828b766483148fb40e1ffa5d557ce9851a535af94965e0dd0988"
        )

        msg = ripple.create_sign_tx_msg(
            {
                "TransactionType": "Payment",
                "Payment": {
                    "Amount": 100000009,
                    "Destination": "rNaqKtKrMSwpwZSzRckPf7S96DkimjkF4H",
                    "DestinationTag": 123456,
                },
                "Flags": 0,
                "Fee": 100,
                "Sequence": 100,
                "LastLedgerSequence": 333111,
            }
        )
        resp = ripple.sign_tx(self.client, parse_path("m/44'/144'/0'/0/2"), msg)
        assert (
            resp.signature.hex()
            == "30450221008770743a472bb2d1c746a53ef131cc17cc118d538ec910ca928d221db4494cf702201e4ef242d6c3bff110c3cc3897a471fed0f5ac10987ea57da63f98dfa01e94df"
        )
        assert (
            resp.serialized_tx.hex()
            == "120000228000000024000000642e0001e240201b00051537614000000005f5e109684000000000000064732103dbed1e77cb91a005e2ec71afbccce5444c9be58276665a3859040f692de8fed2744730450221008770743a472bb2d1c746a53ef131cc17cc118d538ec910ca928d221db4494cf702201e4ef242d6c3bff110c3cc3897a471fed0f5ac10987ea57da63f98dfa01e94df8114bdf86f3ae715ba346b7772ea0e133f48828b766483148fb40e1ffa5d557ce9851a535af94965e0dd0988"
        )

    def test_ripple_sign_invalid_fee(self):
        self.setup_mnemonic_allallall()

        msg = ripple.create_sign_tx_msg(
            {
                "TransactionType": "Payment",
                "Payment": {
                    "Amount": 1,
                    "Destination": "rNaqKtKrMSwpwZSzRckPf7S96DkimjkF4H",
                },
                "Flags": 1,
                "Fee": 1,
                "Sequence": 1,
            }
        )
        with pytest.raises(CallException) as exc:
            ripple.sign_tx(self.client, parse_path("m/44'/144'/0'/0/2"), msg)
        assert exc.value.args[0] == messages.FailureType.ProcessError
        assert exc.value.args[1].endswith(
            "Fee must be in the range of 10 to 10,000 drops"
        )



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .MultisigRedeemScriptType import MultisigRedeemScriptType

if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None  # type: ignore


class TxOutputType(p.MessageType):

    def __init__(
        self,
        address: str = None,
        address_n: List[int] = None,
        amount: int = None,
        script_type: int = None,
        multisig: MultisigRedeemScriptType = None,
        op_return_data: bytes = None,
        decred_script_version: int = None,
        block_hash_bip115: bytes = None,
        block_height_bip115: int = None,
    ) -> None:
        self.address = address
        self.address_n = address_n if address_n is not None else []
        self.amount = amount
        self.script_type = script_type
        self.multisig = multisig
        self.op_return_data = op_return_data
        self.decred_script_version = decred_script_version
        self.block_hash_bip115 = block_hash_bip115
        self.block_height_bip115 = block_height_bip115

    @classmethod
    def get_fields(cls):
        return {
            1: ('address', p.UnicodeType, 0),
            2: ('address_n', p.UVarintType, p.FLAG_REPEATED),
            3: ('amount', p.UVarintType, 0),  # required
            4: ('script_type', p.UVarintType, 0),  # required
            5: ('multisig', MultisigRedeemScriptType, 0),
            6: ('op_return_data', p.BytesType, 0),
            7: ('decred_script_version', p.UVarintType, 0),
            8: ('block_hash_bip115', p.BytesType, 0),
            9: ('block_height_bip115', p.UVarintType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .MoneroTransactionRsigData import MoneroTransactionRsigData

if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None  # type: ignore


class MoneroTransactionInitAck(p.MessageType):
    MESSAGE_WIRE_TYPE = 502

    def __init__(
        self,
        hmacs: List[bytes] = None,
        rsig_data: MoneroTransactionRsigData = None,
    ) -> None:
        self.hmacs = hmacs if hmacs is not None else []
        self.rsig_data = rsig_data

    @classmethod
    def get_fields(cls):
        return {
            1: ('hmacs', p.BytesType, p.FLAG_REPEATED),
            2: ('rsig_data', MoneroTransactionRsigData, 0),
        }



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

from trezorlib import btc, messages as proto

from ..support import ckd_public as bip32
from .common import TrezorTest


class TestMsgGetaddressShow(TrezorTest):
    def test_show(self):
        self.setup_mnemonic_nopin_nopassphrase()
        assert (
            btc.get_address(self.client, "Bitcoin", [1], show_display=True)
            == "1CK7SJdcb8z9HuvVft3D91HLpLC6KSsGb"
        )
        assert (
            btc.get_address(self.client, "Bitcoin", [2], show_display=True)
            == "15AeAhtNJNKyowK8qPHwgpXkhsokzLtUpG"
        )
        assert (
            btc.get_address(self.client, "Bitcoin", [3], show_display=True)
            == "1CmzyJp9w3NafXMSEFH4SLYUPAVCSUrrJ5"
        )

    def test_show_multisig_3(self):
        self.setup_mnemonic_nopin_nopassphrase()

        node = bip32.deserialize(
            "xpub661MyMwAqRbcF1zGijBb2K6x9YiJPh58xpcCeLvTxMX6spkY3PcpJ4ABcCyWfskq5DDxM3e6Ez5ePCqG5bnPUXR4wL8TZWyoDaUdiWW7bKy"
        )
        multisig = proto.MultisigRedeemScriptType(
            pubkeys=[
                proto.HDNodePathType(node=node, address_n=[1]),
                proto.HDNodePathType(node=node, address_n=[2]),
                proto.HDNodePathType(node=node, address_n=[3]),
            ],
            signatures=[b"", b"", b""],
            m=2,
        )

        for i in [1, 2, 3]:
            assert (
                btc.get_address(
                    self.client, "Bitcoin", [i], show_display=True, multisig=multisig
                )
                == "3E7GDtuHqnqPmDgwH59pVC7AvySiSkbibz"
            )

    def test_show_multisig_15(self):
        self.setup_mnemonic_nopin_nopassphrase()

        node = bip32.deserialize(
            "xpub661MyMwAqRbcF1zGijBb2K6x9YiJPh58xpcCeLvTxMX6spkY3PcpJ4ABcCyWfskq5DDxM3e6Ez5ePCqG5bnPUXR4wL8TZWyoDaUdiWW7bKy"
        )

        pubs = []
        for x in range(15):
            pubs.append(proto.HDNodePathType(node=node, address_n=[x]))

        multisig = proto.MultisigRedeemScriptType(
            pubkeys=pubs, signatures=[b""] * 15, m=15
        )

        for i in range(15):
            assert (
                btc.get_address(
                    self.client, "Bitcoin", [i], show_display=True, multisig=multisig
                )
                == "3QaKF8zobqcqY8aS6nxCD5ZYdiRfL3RCmU"
            )



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .HDNodeType import HDNodeType


class DebugLinkState(p.MessageType):
    MESSAGE_WIRE_TYPE = 102

    def __init__(
        self,
        layout: bytes = None,
        pin: str = None,
        matrix: str = None,
        mnemonic_secret: bytes = None,
        node: HDNodeType = None,
        passphrase_protection: bool = None,
        reset_word: str = None,
        reset_entropy: bytes = None,
        recovery_fake_word: str = None,
        recovery_word_pos: int = None,
        reset_word_pos: int = None,
        mnemonic_type: int = None,
    ) -> None:
        self.layout = layout
        self.pin = pin
        self.matrix = matrix
        self.mnemonic_secret = mnemonic_secret
        self.node = node
        self.passphrase_protection = passphrase_protection
        self.reset_word = reset_word
        self.reset_entropy = reset_entropy
        self.recovery_fake_word = recovery_fake_word
        self.recovery_word_pos = recovery_word_pos
        self.reset_word_pos = reset_word_pos
        self.mnemonic_type = mnemonic_type

    @classmethod
    def get_fields(cls):
        return {
            1: ('layout', p.BytesType, 0),
            2: ('pin', p.UnicodeType, 0),
            3: ('matrix', p.UnicodeType, 0),
            4: ('mnemonic_secret', p.BytesType, 0),
            5: ('node', HDNodeType, 0),
            6: ('passphrase_protection', p.BoolType, 0),
            7: ('reset_word', p.UnicodeType, 0),
            8: ('reset_entropy', p.BytesType, 0),
            9: ('recovery_fake_word', p.UnicodeType, 0),
            10: ('recovery_word_pos', p.UVarintType, 0),
            11: ('reset_word_pos', p.UVarintType, 0),
            12: ('mnemonic_type', p.UVarintType, 0),
        }



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

from trezorlib import btc, messages as proto
from trezorlib.tools import parse_path

from ..support import ckd_public as bip32
from .common import TrezorTest


class TestMsgGetaddressSegwitNative(TrezorTest):
    def test_show_segwit(self):
        self.setup_mnemonic_allallall()
        assert (
            btc.get_address(
                self.client,
                "Testnet",
                parse_path("49'/1'/0'/0/0"),
                True,
                None,
                script_type=proto.InputScriptType.SPENDWITNESS,
            )
            == "tb1qqzv60m9ajw8drqulta4ld4gfx0rdh82un5s65s"
        )
        assert (
            btc.get_address(
                self.client,
                "Testnet",
                parse_path("49'/1'/0'/1/0"),
                False,
                None,
                script_type=proto.InputScriptType.SPENDWITNESS,
            )
            == "tb1q694ccp5qcc0udmfwgp692u2s2hjpq5h407urtu"
        )
        assert (
            btc.get_address(
                self.client,
                "Testnet",
                parse_path("44'/1'/0'/0/0"),
                False,
                None,
                script_type=proto.InputScriptType.SPENDWITNESS,
            )
            == "tb1q54un3q39sf7e7tlfq99d6ezys7qgc62a6rxllc"
        )
        assert (
            btc.get_address(
                self.client,
                "Testnet",
                parse_path("44'/1'/0'/0/0"),
                False,
                None,
                script_type=proto.InputScriptType.SPENDADDRESS,
            )
            == "mvbu1Gdy8SUjTenqerxUaZyYjmveZvt33q"
        )
        assert (
            btc.get_address(
                self.client,
                "Groestlcoin",
                parse_path("84'/17'/0'/0/0"),
                False,
                None,
                script_type=proto.InputScriptType.SPENDWITNESS,
            )
            == "grs1qw4teyraux2s77nhjdwh9ar8rl9dt7zww8r6lne"
        )

    def test_show_multisig_3(self):
        self.setup_mnemonic_allallall()
        nodes = [
            btc.get_public_node(self.client, parse_path("999'/1'/%d'" % index))
            for index in range(1, 4)
        ]
        multisig1 = proto.MultisigRedeemScriptType(
            nodes=[bip32.deserialize(n.xpub) for n in nodes],
            address_n=[2, 0],
            signatures=[b"", b"", b""],
            m=2,
        )
        multisig2 = proto.MultisigRedeemScriptType(
            nodes=[bip32.deserialize(n.xpub) for n in nodes],
            address_n=[2, 1],
            signatures=[b"", b"", b""],
            m=2,
        )
        for i in [1, 2, 3]:
            assert (
                btc.get_address(
                    self.client,
                    "Testnet",
                    parse_path("999'/1'/%d'/2/1" % i),
                    False,
                    multisig2,
                    script_type=proto.InputScriptType.SPENDWITNESS,
                )
                == "tb1qch62pf820spe9mlq49ns5uexfnl6jzcezp7d328fw58lj0rhlhasge9hzy"
            )
            assert (
                btc.get_address(
                    self.client,
                    "Testnet",
                    parse_path("999'/1'/%d'/2/0" % i),
                    False,
                    multisig1,
                    script_type=proto.InputScriptType.SPENDWITNESS,
                )
                == "tb1qr6xa5v60zyt3ry9nmfew2fk5g9y3gerkjeu6xxdz7qga5kknz2ssld9z2z"
            )



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .MoneroRingCtSig import MoneroRingCtSig


class MoneroTransactionAllOutSetAck(p.MessageType):
    MESSAGE_WIRE_TYPE = 514

    def __init__(
        self,
        extra: bytes = None,
        tx_prefix_hash: bytes = None,
        rv: MoneroRingCtSig = None,
        full_message_hash: bytes = None,
    ) -> None:
        self.extra = extra
        self.tx_prefix_hash = tx_prefix_hash
        self.rv = rv
        self.full_message_hash = full_message_hash

    @classmethod
    def get_fields(cls):
        return {
            1: ('extra', p.BytesType, 0),
            2: ('tx_prefix_hash', p.BytesType, 0),
            4: ('rv', MoneroRingCtSig, 0),
            5: ('full_message_hash', p.BytesType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class PinMatrixRequest(p.MessageType):
    MESSAGE_WIRE_TYPE = 18

    def __init__(
        self,
        type: int = None,
    ) -> None:
        self.type = type

    @classmethod
    def get_fields(cls):
        return {
            1: ('type', p.UVarintType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None  # type: ignore


class StellarSignTx(p.MessageType):
    MESSAGE_WIRE_TYPE = 202

    def __init__(
        self,
        address_n: List[int] = None,
        network_passphrase: str = None,
        source_account: str = None,
        fee: int = None,
        sequence_number: int = None,
        timebounds_start: int = None,
        timebounds_end: int = None,
        memo_type: int = None,
        memo_text: str = None,
        memo_id: int = None,
        memo_hash: bytes = None,
        num_operations: int = None,
    ) -> None:
        self.address_n = address_n if address_n is not None else []
        self.network_passphrase = network_passphrase
        self.source_account = source_account
        self.fee = fee
        self.sequence_number = sequence_number
        self.timebounds_start = timebounds_start
        self.timebounds_end = timebounds_end
        self.memo_type = memo_type
        self.memo_text = memo_text
        self.memo_id = memo_id
        self.memo_hash = memo_hash
        self.num_operations = num_operations

    @classmethod
    def get_fields(cls):
        return {
            2: ('address_n', p.UVarintType, p.FLAG_REPEATED),
            3: ('network_passphrase', p.UnicodeType, 0),
            4: ('source_account', p.UnicodeType, 0),
            5: ('fee', p.UVarintType, 0),
            6: ('sequence_number', p.UVarintType, 0),
            8: ('timebounds_start', p.UVarintType, 0),
            9: ('timebounds_end', p.UVarintType, 0),
            10: ('memo_type', p.UVarintType, 0),
            11: ('memo_text', p.UnicodeType, 0),
            12: ('memo_id', p.UVarintType, 0),
            13: ('memo_hash', p.BytesType, 0),
            14: ('num_operations', p.UVarintType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .TxRequestDetailsType import TxRequestDetailsType
from .TxRequestSerializedType import TxRequestSerializedType


class TxRequest(p.MessageType):
    MESSAGE_WIRE_TYPE = 21

    def __init__(
        self,
        request_type: int = None,
        details: TxRequestDetailsType = None,
        serialized: TxRequestSerializedType = None,
    ) -> None:
        self.request_type = request_type
        self.details = details
        self.serialized = serialized

    @classmethod
    def get_fields(cls):
        return {
            1: ('request_type', p.UVarintType, 0),
            2: ('details', TxRequestDetailsType, 0),
            3: ('serialized', TxRequestSerializedType, 0),
        }



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import pytest

from trezorlib import debuglink
from trezorlib.ripple import get_address
from trezorlib.tools import parse_path

from .common import TrezorTest


@pytest.mark.ripple
@pytest.mark.skip_t1  # T1 support is not planned
class TestMsgRippleGetAddress(TrezorTest):
    def test_ripple_get_address(self):
        # data from https://iancoleman.io/bip39/#english
        self.setup_mnemonic_allallall()

        address = get_address(self.client, parse_path("m/44'/144'/0'/0/0"))
        assert address == "rNaqKtKrMSwpwZSzRckPf7S96DkimjkF4H"
        address = get_address(self.client, parse_path("m/44'/144'/0'/0/1"))
        assert address == "rBKz5MC2iXdoS3XgnNSYmF69K1Yo4NS3Ws"
        address = get_address(self.client, parse_path("m/44'/144'/1'/0/0"))
        assert address == "rJX2KwzaLJDyFhhtXKi3htaLfaUH2tptEX"

    def test_ripple_get_address_other(self):
        # data from https://github.com/you21979/node-ripple-bip32/blob/master/test/test.js
        debuglink.load_device_by_mnemonic(
            self.client,
            mnemonic="armed bundle pudding lazy strategy impulse where identify submit weekend physical antenna flight social acoustic absurd whip snack decide blur unfold fiction pumpkin athlete",
            pin="",
            passphrase_protection=False,
            label="test",
            language="english",
        )
        address = get_address(self.client, parse_path("m/44'/144'/0'/0/0"))
        assert address == "r4ocGE47gm4G4LkA9mriVHQqzpMLBTgnTY"
        address = get_address(self.client, parse_path("m/44'/144'/0'/0/1"))
        assert address == "rUt9ULSrUvfCmke8HTFU1szbmFpWzVbBXW"



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

from trezorlib import debuglink, device
from trezorlib.messages.PassphraseSourceType import HOST as PASSPHRASE_ON_HOST

from . import conftest


class TrezorTest:
    # fmt: off
    #                1      2     3    4      5      6      7     8      9    10    11    12
    mnemonic12 = "alcohol woman abuse must during monitor noble actual mixed trade anger aisle"
    mnemonic18 = "owner little vague addict embark decide pink prosper true fork panda embody mixture exchange choose canoe electric jewel"
    mnemonic24 = "dignity pass list indicate nasty swamp pool script soccer toe leaf photo multiply desk host tomato cradle drill spread actor shine dismiss champion exotic"
    mnemonic_all = " ".join(["all"] * 12)
    # fmt: on

    pin4 = "1234"
    pin6 = "789456"
    pin8 = "45678978"

    def setup_method(self, method):
        self.client = conftest.get_device()
        # self.client.set_buttonwait(3)

        device.wipe(self.client)
        self.client.open()

    def teardown_method(self, method):
        self.client.close()

    def _setup_mnemonic(self, mnemonic=None, pin="", passphrase=False, lock=True):
        if mnemonic is None:
            mnemonic = TrezorTest.mnemonic12
        debuglink.load_device_by_mnemonic(
            self.client,
            mnemonic=mnemonic,
            pin=pin,
            passphrase_protection=passphrase,
            label="test",
            language="english",
        )
        if conftest.TREZOR_VERSION == 1 and lock:
            # remove cached PIN (introduced via load_device)
            self.client.clear_session()
        if conftest.TREZOR_VERSION > 1 and passphrase:
            device.apply_settings(self.client, passphrase_source=PASSPHRASE_ON_HOST)

    def setup_mnemonic_allallall(self, lock=True):
        self._setup_mnemonic(mnemonic=TrezorTest.mnemonic_all, lock=lock)

    def setup_mnemonic_nopin_nopassphrase(self, lock=True):
        self._setup_mnemonic(lock=lock)

    def setup_mnemonic_nopin_passphrase(self, lock=True):
        self._setup_mnemonic(passphrase=True, lock=lock)

    def setup_mnemonic_pin_nopassphrase(self, lock=True):
        self._setup_mnemonic(pin=TrezorTest.pin4, lock=lock)

    def setup_mnemonic_pin_passphrase(self, lock=True):
        self._setup_mnemonic(pin=TrezorTest.pin4, passphrase=True, lock=lock)


def generate_entropy(strength, internal_entropy, external_entropy):
    """
    strength - length of produced seed. One of 128, 192, 256
    random - binary stream of random data from external HRNG
    """
    import hashlib

    if strength not in (128, 192, 256):
        raise ValueError("Invalid strength")

    if not internal_entropy:
        raise ValueError("Internal entropy is not provided")

    if len(internal_entropy) < 32:
        raise ValueError("Internal entropy too short")

    if not external_entropy:
        raise ValueError("External entropy is not provided")

    if len(external_entropy) < 32:
        raise ValueError("External entropy too short")

    entropy = hashlib.sha256(internal_entropy + external_entropy).digest()
    entropy_stripped = entropy[: strength // 8]

    if len(entropy_stripped) * 8 != strength:
        raise ValueError("Entropy length mismatch")

    return entropy_stripped



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .IdentityType import IdentityType


class SignIdentity(p.MessageType):
    MESSAGE_WIRE_TYPE = 53

    def __init__(
        self,
        identity: IdentityType = None,
        challenge_hidden: bytes = None,
        challenge_visual: str = None,
        ecdsa_curve_name: str = None,
    ) -> None:
        self.identity = identity
        self.challenge_hidden = challenge_hidden
        self.challenge_visual = challenge_visual
        self.ecdsa_curve_name = ecdsa_curve_name

    @classmethod
    def get_fields(cls):
        return {
            1: ('identity', IdentityType, 0),
            2: ('challenge_hidden', p.BytesType, 0),
            3: ('challenge_visual', p.UnicodeType, 0),
            4: ('ecdsa_curve_name', p.UnicodeType, 0),
        }
